# On-The-Fly Workload Partitioning for Integrated CPU/GPU Architectures

## Overview

This artifact demonstrates the effectiveness and the benefits of the presented on-the-fly workload partitioning technique for irregular data-parallel workloads on integrated CPU/GPU architectures.
The technique requires no offline analysis or training of the application or the input data. For an OpenCL kernel, a source-to-source compiler creates profiling code that allows the runtime system to collect information about the computational load of the work-items.
Based on this profile information, the workload is reshaped such a way that work-items with loop iteration counts above a dynamically determined threshold are executed on the CPU cores while the GPU only executes work-items with a similar computational load.

The framework, named ArmCL (ARM OpenCL framework), is part of the ARM (Advanced Resource Management) research project by the Computer Systems and Platforms Laboratory at Seoul National University and the Laboratory for Software Technology at ETH Zurich and consists of three components.

**compiler (prep)**: the preprocessor that performs code generation for the on-the-fly workload partitioning techniuqe.

**runtime (lib)**: the dynamically loadable library that performs the on-the-fly workload partitioning by leveraging the code generation.

**benchmark (bench)**: it contains four benchmark applications and some scripts to help experiments.

## Target Architectures and Dependencies

The framework assumes an integrated CPU/GPU architecture where the CPU and the GPU share the same off-chip DRAM and where OpenCL >v1.2. is installed.
Internally, the framework uses the installed OpenCL libraries and thus has no software dependencies.


### The AMD Kaveri System

CPUS: 4x AMD A10-7850K Radeon R7

Memory: 16GB

g++ version: 4.8.4

Kernel version: 3.19.0 (Ubuntu)

### The Intel Skylake System

CPUS: 4x Intel(R) Core(TM) i5-6600 CPU 3.30GHz

Memory: 16GB

g++ version: 5.4.0

Kernel version: 4.7.0 (Ubuntu)

## Login Credentials

Access to the AMD Kaveri and the Intel Skylake system used in the evaluation of the papar may be provided upon request. Send your request to bernhard@csap.snu.ac.kr.

### The AMD Kaveri System
```
$ ssh <obtain login credentials>
```

### The Intel Skylake System
```
$ ssh <obtain login credentials>
$ sh skylake.sh
```

## Installation

### The dynamically loadable library
The code for the workload partitioning library is located in the *lib* directory. Here the core part of the on-the-fly workload partitioning is provided as an object file compiled for each target machine. The necesasry code for automatic CPU threading and coarse-grained workload partitioning is provided in source-code form.

Install the ArmCL library as follow.

```
$ cd lib/
```

For the AMD Kaveri system,
```
$ make -f Makefile.amd
```

For the Intel Skylake system,
```
$ make -f Makefile.intel
```

### The preprocessor

As of now the preprocessor is delivered as a binary program compiled for the AMD Kaveri APU system and the Intel i5-6600 processor described above.

## Usage: Individual Runs

To run an OpenCL application with the ArmCL library, two environment variables need to be set: 1. *ARMCL\_PATH* to indicate the location of the *.so* file of the library, 2. *PATH* to indicate the location of the preprocessor.

The following script automates the setting of these variables.
```
$ source ./env.sh 
```

Prepare data sets.
```
$ sh prepare_dataset.py
```

### Native (unmodified) GPU execution
```
$ ./a.out
```

### CPU multithreading
This version executes the application using as many threads as there are CPU cores in the system.
```
$ LD_PRELOAD=$ARMCL_PATH/libArmCL.so METHOD=CPU ./a.out
```

### Coarse-grained workgroup partitioning
This version executes the application using both the CPU and the GPU by dividing work-groups without any workload reshaping. The ratio of work division between the CPU and the GPU needs to be specified by setting the *CPU\_RATIO* and *GPU\_RATIO* environment variables.
```
$ LD_PRELOAD=$ARMCL_PATH/libArmCL.so METHOD=WorkGroupPartitioning CPU_RATIO=20 GPU_TAIO=80 ./a.out
```

### Manual workload partitioning using workload reshaping
This version executes the application using both the CPU and the GPU based on workload reshaping (the Shift-H method described in the paper). Also here, the ratio of work division between the CPU and the GPU needs to be specified by defining the *CPU\_RATIO* and *GPU\_RATIO* environment variables.
```
$ LD_PRELOAD=$ARMCL_PATH/libArmCL.so METHOD=Manual CPU_RATIO=20 GPU_TAIO=80 ./a.out
```

### On-the-fly workload partitioning
This version executes the application using both the CPU and the GPU using automatic workload reshaping through the presented on-the-fly workload partitioning technique. The threshold values (the CPU and GPU ratio) are determined automatically at runtime.
```
$ LD_PRELOAD=$ARMCL_PATH/libArmCL.so METHOD=OnTheFly ./a.out
```

## Benchmarks: Reproducing the Results

### Preparation
In the *bench* directory, benchmark applications (*PageRank*, *SpMV*, *GraphColoring*, *BFS*) and helper scripts are provided that allow a reproduction of the results in the paper.
Before executing the benchmarks, the necessary datasets and reference outputs need to be downloaded using the following scripts.
```
$ sh prepare_dataset.py
$ sh prepare_refoutput.py
```

In each directory we provide five run scripts to reproduce the results in the paper.
Compile each application as follows.

For the AMD Kaveri system,
```
$ make -f Makefile.amd
```

For the Intel Skylake system,
```
$ make -f Makefile.intel
```

### Native (unmodified) applications for each dataset
```
$ python run_gpu.py
```
### Executing applications using CPU threading for each dataset
```
$ python run_cpu.py
```
### Executing applications using coarse-grained workgroup partitioning (for changing CPU ratio from 5% to 95%) for each dataset
```
$ python run_group.py
```
### Executing applications using workload reshaping (Shift-H method; refer to the paper) (for changing CPU ratio from 1% to 99%) for each dataset
```
$ python run_oracle.py
```
note that the execution time of *run_oracle.py* is in the order of several hours to days because the script tests every possible workload division between the CPU and the GPU to find the oracle partitioning point.
You can skip this process and copy the provided data from the *oracle-data-kaveri* (generated on our AMD Kaveri machine) or *oralce-data-skylake* (generated on our Intel Skylake machine) directories. For example, *$ cp -r oracle-data-kaveri/oracle csv/oracle*.

### Executing applications using the on-the-fly workload partitioning technique.
```
$ python run_on_the_fly.py
```
### Analysis
The following script is used to analyze the numbers from the above scripts. The scripts then produces csv files (e.g. *on_the_fly_normalized.csv*) showing the normalized execution time in comparison to the CPU threading version.
```
$ python run_analysis.py
```

Once the *csv* files have been generated, *gen_plot.gnu* generates a bar graph in eps format.
```
$ ./gen_plot.gnu
or
$ gnuplot ./gen_plot.gnu
```

## Publication 
Younghyun Cho, Florian Negele, Seohong Park, Bernhard Egger, and Thomas R. Gross. "On-The-Fly Workload Partitioning for Integrated CPU/GPU Architectures", to appear in Proceedings of the International Conference on Parallel Architectures and Compilation Techniques (PACT 2018), Limassol, Cyprus, 2018.
