// The original application code is obtained from GraphBIG benchmark suite
// https://github.com/graphbig
//
// The CUDA version of GraphBIG applications was ported into OpenCL by
// Younghyun Cho <younghyun@csap.snu.ac.kr>
// Seohong Park <artberryx@gmail.com>

#include <CL/cl.h>
#include <vector>
#include <string>
#include <fstream>
#include <thread>
#include <atomic>
#include <malloc.h>
#include "common.h"
#include "def.h"
#include "openG.h"
#include "oclGraph.h"
#include <time.h>

using namespace std;

class vertex_property
{
    public:
        vertex_property():value(0){}
        vertex_property(unsigned int x):value(x){}

        unsigned int value;
};
class edge_property
{
    public:
        edge_property():value(0){}
        edge_property(unsigned int x):value(x){}

        unsigned int value;
};

typedef openG::extGraph<vertex_property, edge_property> graph_t;
typedef graph_t::vertex_iterator    vertex_iterator;
typedef graph_t::edge_iterator      edge_iterator;

#define CHECK_ERROR(err)                                   \
    if (err != CL_SUCCESS) {                                 \
        printf("[%s:%d] error %d\n", __FILE__, __LINE__, err); \
        exit(EXIT_FAILURE);                                    \
    }

#define CHECK_FINISH() { cl_int ecode = clFinish(command_queue); \
    CHECK_ERROR(ecode); }

static long long getTimestamp(void)
{
    static __thread struct timeval tv;
    gettimeofday(&tv, NULL);
    return 1000000LL * tv.tv_sec + tv.tv_usec;
}

//==============================================================//

#define SEED 123

void output(unsigned int* vproplist, int vertex_cnt)
{
    cout<<"Graph Coloring Results:\n";
    for (int i = 0; i < vertex_cnt; i++) {
        cout << "== vertex " << i << ": color " << vproplist[i] << endl;
    }
}

typedef unsigned int uint;

int ocl_graph_coloring(uint64_t * vertexlist, uint64_t * edgelist,
        unsigned int vertex_cnt, unsigned int edge_cnt)
{
    cl_uint num_platforms;
    cl_platform_id* platforms;
    size_t platform_name_size;
    char* platform_name;

    cl_platform_id target_ocl_platform = NULL;
    cl_device_id device;
    cl_context context;
    cl_command_queue command_queue;
    cl_program program;
    cl_kernel initialize, kernel_step;

    cl_int err;

    err = clGetPlatformIDs(0, NULL, &num_platforms);
    CHECK_ERROR(err);
    printf("%d platforms are detected.\n", (int) num_platforms);

    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id) * num_platforms);
    err = clGetPlatformIDs(num_platforms, platforms, NULL);
    CHECK_ERROR(err);

    for (int i = 0; i < num_platforms; i++) {
        err = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 0, NULL,
                &platform_name_size);
        CHECK_ERROR(err);

        platform_name = (char*)malloc(sizeof(char) * platform_name_size);
        err = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, platform_name_size,
                platform_name, NULL);

        printf("Platform: %d: %s\n", i, platform_name);

        if (strcmp(platform_name, PLATFORM_NAME) == 0)
            target_ocl_platform = platforms[i];
        //target_ocl_platform = platforms[i];
        free(platform_name);
    }

    if (target_ocl_platform == NULL) {
        printf("target platform is not found.\n");
        exit(EXIT_FAILURE);
    }

    err = clGetDeviceIDs(target_ocl_platform, CL_DEVICE_TYPE_GPU, 1, &device,
            NULL);
    CHECK_ERROR(err);

    context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);
    CHECK_ERROR(err);

    command_queue = clCreateCommandQueue(context, device, 0, &err);
    CHECK_ERROR(err);

    FILE* fp = fopen("./graph_coloring.cl", "r");
    if (!fp) {
        fprintf(stderr, "Failed to load kernel.\n");
        exit(-1);
    }

    char* source_str = (char*)malloc(0x100000);
    size_t source_size = fread(source_str, 1, 0x100000, fp);
    fclose(fp);

    program = clCreateProgramWithSource(context,
            1,
            (const char**)&source_str,
            (const size_t*)&source_size,
            &err);
    CHECK_ERROR(err);

    err = clBuildProgram(program, 1, &device, NULL, NULL, NULL);
    if(err != CL_SUCCESS) // detailed_CHECK_ERROR
    {
        printf("[%s:%d] error %d\n", __FILE__, __LINE__, err);
        if (err == CL_BUILD_PROGRAM_FAILURE) {
            size_t log_size;
            clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, 0, NULL, &log_size);
            char *log = (char *) malloc(log_size);
            clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG, log_size, log, NULL);
            printf("%s\n", log);
        }
        exit(EXIT_FAILURE);
    }

    initialize = clCreateKernel(program, "initialize", &err);
    CHECK_ERROR(err);
    kernel_step = clCreateKernel(program, "kernel_step", &err);
    CHECK_ERROR(err);

    // work item padding for GPU allocation unit
    size_t wg_size = 64; // preferred work group size
    size_t total_items = (vertex_cnt+(wg_size-1))/wg_size*wg_size;

    // prepare graph struct
    unsigned int* vertexlist_ = (unsigned int*)malloc((vertex_cnt+1) * sizeof(unsigned int));
    unsigned int* edgelist_ = (unsigned int*)malloc(edge_cnt * sizeof(unsigned int));
    for (int i = 0; i < vertex_cnt+1; i++)
        vertexlist_[i] = (unsigned int)(vertexlist[i]);
    for (int i = 0; i < edge_cnt; i++)
        edgelist_[i] = (unsigned int)(edgelist[i]);
    unsigned int* device_vpl = (unsigned int*)memalign(4096, total_items * sizeof(unsigned int));
    for (int i = 0; i < total_items; i++)
        device_vpl[i] = 0;
    unsigned int* device_rand = (unsigned int*)memalign(4096, total_items * sizeof(unsigned int));
    for (int i = 0; i < total_items; i++)
        device_rand[i] = 0;
    unsigned int* device_over = (unsigned int*)memalign(4096, 1024*sizeof(unsigned int));
    for (int i = 0; i < 1024; i++)
        device_over[i] = 0;

    cl_mem device_vertexlist_mem = clCreateBuffer(context,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            (vertex_cnt+1)*sizeof(unsigned int),
            vertexlist_,
            &err);
    CHECK_ERROR(err);
    cl_mem device_edgelist_mem = clCreateBuffer(context,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            edge_cnt*sizeof(unsigned int),
            edgelist_,
            &err);
    CHECK_ERROR(err);
    cl_mem device_vpl_mem = clCreateBuffer(context,
            CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR,
            (total_items)*sizeof(unsigned int),
            device_vpl,
            &err);
    CHECK_ERROR(err);
    cl_mem device_rand_mem = clCreateBuffer(context,
            CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR,
            (total_items)*sizeof(unsigned int),
            device_rand,
            &err);
    CHECK_ERROR(err);
    cl_mem device_over_mem = clCreateBuffer(context,
            CL_MEM_READ_WRITE|CL_MEM_USE_HOST_PTR,
            (1024)*sizeof(unsigned int),
            device_over,
            &err);
    CHECK_ERROR(err);

    err = clSetKernelArg(initialize, 0, sizeof(unsigned int), (void*)&vertex_cnt);
    err |= clSetKernelArg(initialize, 1, sizeof(cl_mem), (void*)&device_vpl_mem);
    CHECK_ERROR(err);

    err = clSetKernelArg(kernel_step, 0, sizeof(cl_mem), (void*)&device_vertexlist_mem);
    err |= clSetKernelArg(kernel_step, 1, sizeof(cl_mem), (void*)&device_edgelist_mem);
    err |= clSetKernelArg(kernel_step, 2, sizeof(unsigned int), (void*)&vertex_cnt);
    err |= clSetKernelArg(kernel_step, 3, sizeof(cl_mem), (void*)&device_vpl_mem);
    err |= clSetKernelArg(kernel_step, 4, sizeof(cl_mem), (void*)&device_rand_mem);
    err |= clSetKernelArg(kernel_step, 6, sizeof(cl_mem), (void*)&device_over_mem);
    CHECK_ERROR(err);

    size_t index_space_[1] = { total_items };
    size_t local_space_[1] = { wg_size };

    //long long elapsed_time = getTimestamp();

    printf("Initialize\n");
    {
        CHECK_FINISH();
        err = clEnqueueNDRangeKernel(command_queue,
                initialize,
                1,
                NULL,
                index_space_,
                local_space_,
                0,
                NULL,
                NULL);
        CHECK_ERROR(err);
        CHECK_FINISH();
    }

    srand(SEED);
    for(uint i=0;i<vertex_cnt;i++)
    {
        device_rand[i] = rand();
    }

    struct timespec total_start, total_end;
    struct timespec start, end;

    clock_gettime(CLOCK_REALTIME, &total_start);

    int kernel_step_num = 0;
    int stop;
    uint curr = 0;
    do
    {
        stop = 1;

        device_over[0] = stop;

        err = clSetKernelArg(kernel_step, 5, sizeof(uint), (void*)&curr);
        CHECK_ERROR(err);

        clock_gettime(CLOCK_REALTIME, &start);

        err = clEnqueueNDRangeKernel(command_queue,
                kernel_step,
                1,
                NULL,
                index_space_,
                local_space_,
                0,
                NULL,
                NULL);
        CHECK_ERROR(err);
        CHECK_FINISH();

        clock_gettime(CLOCK_REALTIME, &end);

        double elapsed_time = (end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec);
        elapsed_time /= 1000;
    
        fprintf(stdout, "kernel_elapsed_time: %lf\n", elapsed_time);

        stop = device_over[0];

        curr++;
    } while(!stop);

    clock_gettime(CLOCK_REALTIME, &total_end);

    double elapsed_time = (total_end.tv_sec - total_start.tv_sec)*1000000000 + (total_end.tv_nsec - total_start.tv_nsec);
    elapsed_time /= 1000;

    printf("== iteration #: %d\n", curr);

    fprintf(stdout, "total_elapsed_time: %lf\n", elapsed_time);

    cout<<"\nGPU Graph Coloring finish: \n";
    cout<<"== "<<vertex_cnt<<" vertices  "<<edge_cnt<<" edges\n";

    cout << "iteration: " << curr << endl;
    output(device_vpl, vertex_cnt);

    clReleaseMemObject(device_vertexlist_mem);
    clReleaseMemObject(device_edgelist_mem);
    clReleaseMemObject(device_vpl_mem);
    clReleaseMemObject(device_rand_mem);
    clReleaseMemObject(device_over_mem);

    clReleaseKernel(initialize);
    clReleaseKernel(kernel_step);

    clReleaseProgram(program);
    clReleaseCommandQueue(command_queue);
    clReleaseContext(context);
}

int main(int argc, char * argv[])
{
    cout<<"loading data... \n";

#if defined(SMALL)
    string path="../dataset/small";
    size_t vertex_num, edge_num;
    string vfile = path + "/vertex.CSR";
    string efile = path + "/edge.CSR";

    vector<uint64_t> vertexlist, edgelist;

    graph_t::load_CSR_Graph(vfile, efile,
    vertex_num, edge_num,
    vertexlist, edgelist);
#elif defined(REGULAR)
    string path="random generator";
    int i, j;
    size_t vertex_num = 1000000;
    size_t edge_num = 20000000;
    vector<uint64_t> vertexlist(vertex_num+1, 0), edgelist(edge_num, 0);
    vector<vector<int>> v(vertex_num);
    for(i=0;i<edge_num;i++)
    {
        int x, y;
        while(1)
        {
            x=rand()%vertex_num, y=rand()%vertex_num;
            if(x != y) break;
        }
        v[x].push_back(y);
    }
    int inow = 0;
    for(i=0;i<vertex_num;i++)
    {
        vertexlist[i]=inow;
        for(j=0;j<v[i].size();j++) edgelist[inow++]=v[i][j];
    }
    vertexlist[i]=inow;
#else
#if defined(FULLCHIP)
    string path = "../dataset/FullChip/";
    string separator = " ";
#elif defined(SCALE20)
    string path = "../dataset/scale20/";
    string separator = " ";
#elif defined(IN2004)
    string path = "../dataset/in-2004/";
    string separator = " ";
#elif defined(WatsonGene)
    string path = "../dataset/Watson_Gene/";
    string separator = ",";
#elif defined(LDBC)
    string path = "../dataset/LDBC/output-1000k/";
    string separator = "|";
#elif defined(KNOWLEDGE_REPO)
    string path = "../dataset/Knowledge_Repo/";
    string separator = ",";
#else
    string path = argv[1];
    string separator = ",";
#endif
    string vfile = path + "/vertex.csv";
    string efile = path + "/edge.csv";
    vector<uint64_t> vertexlist, edgelist;

    graph_t graph;
    if (graph.load_csv_vertices(vfile, true, separator, 0) == -1)
        return -1;
    if (graph.load_csv_edges(efile, true, separator, 0, 1) == -1)
        return -1;

    size_t vertex_num = graph.num_vertices();
    size_t edge_num = graph.num_edges();
    graph.to_CSR_Graph(vertexlist, edgelist);
#endif

    //================================================//
    ocl_graph_coloring(&(vertexlist[0]), &(edgelist[0]),
            vertexlist.size()-1, edgelist.size());
    //================================================//

    cout<<"==================================================================\n";
    return 0;
}  // end main

