// Benchmark SpMV application for the on-the-fly workload partitioing technique.
//
// Authors:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
// Seohong Park <artberryx@gmail.com>
//
// Copyright (C) Computer Systems and Platforms Laboratory
// http://csap.snu.ac.kr
// All rights reserved.

#include <CL/cl.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <vector>
#include <cstdio>
#include <cstdlib>
#include <thread>
#include <iostream>
#include <algorithm>
#include <time.h>

#define CHECK_ERROR(err)                                   \
	if (err != CL_SUCCESS) {                                 \
		printf("[%s:%d] error %d\n", __FILE__, __LINE__, err); \
		exit(EXIT_FAILURE);                                    \
	}

#define CHECK_FINISH() { cl_int ecode = clFinish(command_queue); \
	CHECK_ERROR(ecode); }

int main(int argc, char* argv[])
{
    cl_uint num_platforms;
    cl_platform_id* platforms;
    size_t platform_name_size;
    char* platform_name;

    cl_platform_id target_ocl_platform = NULL;
    cl_device_id device;
    cl_context context;
    cl_command_queue command_queue;
    cl_program program;
    cl_kernel kernel;

    cl_int err;

    err = clGetPlatformIDs(0, NULL, &num_platforms);
    CHECK_ERROR(err);
    printf("%u platforms are detected.\n", num_platforms);

    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id) * num_platforms);
    err = clGetPlatformIDs(num_platforms, platforms, NULL);
    CHECK_ERROR(err);

    for (int i = 0; i < num_platforms; i++) {
        err = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME, 0, NULL,
                &platform_name_size);
        CHECK_ERROR(err);

        platform_name = (char*)malloc(sizeof(char) * platform_name_size);
        err = clGetPlatformInfo(platforms[i], CL_PLATFORM_NAME,
                platform_name_size, platform_name, NULL);

        printf("Platform: %d: %s (%s)\n", i, platform_name, PLATFORM_NAME);

        if (strcmp(platform_name, PLATFORM_NAME) == 0)
            target_ocl_platform = platforms[i];
        free(platform_name);
    }

    if (target_ocl_platform == NULL) {
        printf("target platform is not found.\n");
        exit(EXIT_FAILURE);
    }

    err = clGetDeviceIDs(target_ocl_platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    CHECK_ERROR(err);

    context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);
    CHECK_ERROR(err);

    command_queue = clCreateCommandQueue(context, device, 0, &err);
    CHECK_ERROR(err);

    FILE* fp = fopen("./SpMV-CSR.cl", "r");
    if (!fp) {
        fprintf(stderr, "Failed to load kernel.\n");
        exit(-1);
    }

    char* source_str = (char*)malloc(0x100000);
    size_t source_size = fread(source_str, 1, 0x100000, fp);
    fclose(fp);

    program = clCreateProgramWithSource(context,
            1,
            (const char**)&source_str,
            (const size_t*)&source_size,
            &err);
    CHECK_ERROR(err);

    err = clBuildProgram(program, 1, &device, NULL, NULL, NULL);
    CHECK_ERROR(err);

    /// path of the input matrix data
#if defined(REGULAR)
    int width = 1000000;
    int height = 1000000;
    int nnz = 1000000 * 50;
#elif defined(FULLCHIP)
    FILE* infile = fopen("../dataset/FullChip/FullChip.mtx", "r");
    int width, height, nnz;
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#elif defined(SCALE20)
    FILE* infile = fopen("../dataset/scale20/matrxtest.txt", "r");
    int width, height, nnz;
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#elif defined(IN2004)
    FILE* infile = fopen("../dataset/in-2004/in-2004.mtx", "r");
    int width, height, nnz;
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#elif defined(WatsonGene)
    FILE* infile = fopen("../dataset/Watson_Gene/edge.csv", "r");
    //int width = 2041302;
    //int height = 2041302;
    int width = 2041344;
    int height = 2041344;
    int nnz = 12203771;

    char src[10];
    char dst[10];
    fscanf(infile, "%s,%s", src, dst);
    printf("%s %s\n", src, dst);
#elif defined(LDBC)
    FILE* infile = fopen("../dataset/LDBC/output-1000k/edge.csv", "r");
    const int width = 1000000;
    const int height = 1000000;
    const int nnz = 28826024;

    char src[10];
    char dst[10];
    fscanf(infile, "%s|%s", src, dst);
    printf("%s %s\n", src, dst);
#elif defined(KNOWLEDGE_REPO)
    FILE* infile = fopen("../dataset/Knowledge_Repo/edge.csv", "r");
    const int width = 138624;
    const int height = 138624;
    const int nnz = 1394825;
    
    char src[10];
    fscanf(infile, "%s", src);
    printf("%s\n", src);
#else
    FILE* infile = fopen(argv[1], "r");
    int width, height, nnz;
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#endif

    size_t mat_size = width;
    size_t num_elements = nnz;

    int numRows = mat_size;

    size_t size_rowstr = (mat_size+1) * sizeof(int);
    size_t size_a = num_elements * sizeof(float);
    size_t size_p = mat_size * sizeof(float);
    size_t size_colidx = num_elements * sizeof(int);
    size_t size_q = mat_size * sizeof(float);

    int* rowstr = (int*)aligned_alloc(4096, size_rowstr);
    float* a = (float*)aligned_alloc(4096, size_a);
    float* p = (float*)aligned_alloc(4096, size_p);
    int* colidx = (int*)aligned_alloc(4096, size_colidx);
    float* q = (float*)aligned_alloc(4096, size_q);

    //// input data initialization
    int* num_elements_each_row = (int*)malloc(mat_size * sizeof(int));
    for (int i = 0; i < mat_size; i++) {
        num_elements_each_row[i] = 0;
    }

    int* cur_element_each_row = (int*)malloc(mat_size * sizeof(int));
    for (int i = 0; i < mat_size; i++) {
        cur_element_each_row[i] = 0;
    }

#if defined(REGULAR)
    for (int i = 0; i < mat_size; i++) {
        num_elements_each_row[i] = 50;
    }
#else
    for (int i = 0; i < nnz; i++) {
        int rowid, colid;
        float t;
        //fscanf(infile, "%d %d %f", &rowid, &colid, &t);
#if defined(FULLCHIP)
        double padding;
        fscanf(infile, "%d %d %lf", &rowid, &colid, &padding);
#elif defined(SCALE20)
        fscanf(infile, "%d %d", &rowid, &colid);
#elif defined(IN2004)
        fscanf(infile, "%d %d", &rowid, &colid);
#elif defined(WatsonGene)
        fscanf(infile, "%d,%d", &rowid, &colid);
#elif defined(LDBC)
        fscanf(infile, "%d|%d", &rowid, &colid);
#elif defined(KNOWLEDGE_REPO)
        fscanf(infile, "%d,%d", &rowid, &colid);
#else
        fscanf(infile, "%d %d", &rowid, &colid);
#endif
        num_elements_each_row[rowid] += 1;
    }
#endif

    int** elements_each_row = (int**)malloc(mat_size * sizeof(int*));
    for (int i = 0; i < mat_size; i++) {
        elements_each_row[i] = (int*)malloc(num_elements_each_row[i] * sizeof(int));
    }

#if defined(REGULAR)
#elif defined(FULLCHIP)
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#elif defined(SCALE20)
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#elif defined(IN2004)
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#elif defined(WatsonGene)
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%s,%s", src, dst);
    printf("%s %s\n", src, dst);
#elif defined(LDBC)
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%s|%s", src, dst);
    printf("%s %s\n", src, dst);
#elif defined(KNOWLEDGE_REPO)
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%s", src);
    printf("%s\n", src);
#else
    fseek(infile, 0, SEEK_SET);
    fscanf(infile, "%d %d %d", &width, &height, &nnz);
#endif

#if defined(REGULAR)
    for (int i = 0; i < mat_size; i++) {
        for (int j = 0; j < 50; j++) {
            elements_each_row[i][j] = j;
        }
    }
#else
    for (int i = 0; i < nnz; i++) {
        int rowid, colid;
        float t;
#if defined(FULLCHIP)
        double padding;
        fscanf(infile, "%d %d %lf", &rowid, &colid, &padding);
#elif defined(SCALE20)
        fscanf(infile, "%d %d", &rowid, &colid);
#elif defined(IN2004)
        fscanf(infile, "%d %d", &rowid, &colid);
#elif defined(WatsonGene)
        fscanf(infile, "%d,%d", &rowid, &colid);
#elif defined(LDBC)
        fscanf(infile, "%d|%d", &rowid, &colid);
#elif defined(KNOWLEDGE_REPO)
        fscanf(infile, "%d,%d", &rowid, &colid);
#else
        fscanf(infile, "%d %d", &rowid, &colid);
#endif

        int cur_element = cur_element_each_row[rowid];
        cur_element_each_row[rowid] += 1;
        elements_each_row[rowid][cur_element] = colid;
    }
#endif

    {
        int idx = 0;
        int i = 0;
        for (i = 0; i < mat_size; i++) {
            rowstr[i] = idx;
            int num_elems = cur_element_each_row[i];
            for (int j = 0; j < num_elems; j++) {
                colidx[idx] = elements_each_row[i][j];
                idx++;
            }
        }
        rowstr[i] = idx;
    }

    for (int i = 0; i < mat_size; i++)
        free(elements_each_row[i]);
    free(elements_each_row);
    free(cur_element_each_row);

    for (int i = 0; i < mat_size; i++) {
        //p[i] = i * 0.9;
        p[i] = i;
        q[i] = 0;
    }

    for (int i = 0; i < num_elements; i++) {
        //a[i] = (float)i * 0.9;
        a[i] = i;
    }

    //// input data initialization is now finished.

    //// additional data buffers for profiling and execution

    size_t wg_size = 64;
    size_t total_items = (mat_size+(wg_size-1))/wg_size*wg_size;
    size_t total_wgs = total_items / wg_size;

    size_t index_space_[1] = { total_items };
    size_t local_space_[1] = { wg_size };

    //// generate OpenCL buffers
	cl_mem m_rowstr = clCreateBuffer(context,
            CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,
            (mat_size+1) * sizeof(int),
            rowstr,
            &err);
	CHECK_ERROR(err);

	cl_mem m_a = clCreateBuffer(context,
            CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,
            size_a,
            a,
            &err);
	CHECK_ERROR(err);

	cl_mem m_p = clCreateBuffer(context,
            CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,
            size_p,
            p,
            &err);
	CHECK_ERROR(err);

	cl_mem m_colidx = clCreateBuffer(context,
            CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,
            size_colidx,
            colidx,
            &err);
	CHECK_ERROR(err);

	cl_mem m_q = clCreateBuffer(context,
            CL_MEM_READ_WRITE | CL_MEM_USE_HOST_PTR,
            size_q,
            q,
            &err);
	CHECK_ERROR(err);

    kernel = clCreateKernel(program, "SpMV", &err);
    CHECK_ERROR(err);

	err = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void*)&m_rowstr);
	err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), (void*)&m_a);
	err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), (void*)&m_p);
	err |= clSetKernelArg(kernel, 3, sizeof(cl_mem), (void*)&m_colidx);
	err |= clSetKernelArg(kernel, 4, sizeof(cl_mem), (void*)&m_q);
    err |= clSetKernelArg(kernel, 5, sizeof(int), (void*)&numRows);
    CHECK_ERROR(err);

    CHECK_FINISH();

    //// execute the OpenCL kernel
    {
        struct timespec start, end;
        clock_gettime(CLOCK_REALTIME, &start);

        err = clEnqueueNDRangeKernel(command_queue,
                kernel,
                1,
                NULL, // offset
                index_space_,
                local_space_,
                0,
                NULL,
                NULL);
        CHECK_ERROR(err);
        CHECK_FINISH(); // gpu synchronization

        clock_gettime(CLOCK_REALTIME, &end);
        double elapsed_time = (end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec);
        elapsed_time /= 1000;

        fprintf(stdout, "kernel_elapsed_time: %lu\n", (unsigned long)elapsed_time);
    }

    {
        FILE* out = fopen("output.txt", "w");
        float* pre_q = (float*)malloc(size_q);

        for (int id = 0; id < mat_size; id++) {
            float suml = 0.0;
            for (int k = rowstr[id]; k < rowstr[id+1]; k++)
                suml = suml + a[k]*p[colidx[k]];
            pre_q[id] = suml;
        }

        for (int id = 0; id < mat_size; id++) {
            fprintf(out, "%d, %f, %f\n", id, q[id], pre_q[id]);
        }

        free(pre_q);
    }

    clReleaseMemObject(m_rowstr);
    clReleaseMemObject(m_a);
    clReleaseMemObject(m_p);
    clReleaseMemObject(m_colidx);
    clReleaseMemObject(m_q);

    clReleaseKernel(kernel);
	clReleaseProgram(program);
	clReleaseCommandQueue(command_queue);
	clReleaseContext(context);

    free(rowstr);
    free(a);
    free(p);
    free(colidx);
    free(q);

    return 0;
}
