//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) Computer Systems and Platforms Laboratory
// http://csap.snu.ac.kr
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLCommandQueue.h"
#include <sys/time.h>
#include <string.h>
#include <algorithm>

static long long get_time_stamp(void)
{
    static __thread struct timeval tv;
    gettimeofday(&tv, NULL);
    return 1000000LL * tv.tv_sec + tv.tv_usec;
}

ArmCLCommandQueue::ArmCLCommandQueue(cl_command_queue command_queue, ArmCLContext* armcl_context)
    : command_queue_(command_queue)
{
    context_ = armcl_context->context();
    cpu_context_ = armcl_context->cpu_context();

    num_gpu_compute_units_ = armcl_context->num_gpu_compute_units();
    num_cpu_compute_units_ = armcl_context->num_cpu_compute_units();

    //device_ = armcl_context->device_id();
    cpu_device_ = armcl_context->cpu_device_id();

    armcl_context_ = armcl_context;

    PrepareOpenCLCalls();

    // prepare remap buffers
    in_order_map_ = (int*)aligned_alloc(4096, sizeof(int) * max_work_items);
    for (int i = 0; i < max_work_items; i++)
        in_order_map_[i] = i;

    remap_ = (int*)aligned_alloc(4096, sizeof(int) * max_work_items);
    for (int i = 0; i < max_work_items; i++)
        remap_[i] = i;

    array_gpu_ = (int*)aligned_alloc(4096, sizeof(int) * max_work_items);
    for (int i = 0;i < max_work_items; i++)
        array_gpu_[i] = 0;

    array_cpu_ = (int*)aligned_alloc(4096, sizeof(int) * max_work_items);
    for (int i = 0; i < max_work_items; i++)
        array_cpu_[i] = 0;

    loop_size_ = (int*)aligned_alloc(4096, sizeof(int) * max_work_items);
    for (int i = 0; i < max_work_items; i++)
        loop_size_[i] = 0;

    cl_int err;

    cpu_command_queue_ = clCreateCommandQueue_p_(cpu_context_,
            cpu_device_, 0, &err);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: "
            << " clCreateCommandQueue_p_ failed for cpu device"
            << std::endl;
    }
}

ArmCLCommandQueue::~ArmCLCommandQueue()
{
}

void ArmCLCommandQueue::PrepareOpenCLCalls()
{
    char* error;

    clCreateCommandQueue_p_ = (cl_command_queue CL_API_CALL (*) (
                cl_context context,
                cl_device_id device,
                cl_command_queue_properties properties,
                cl_int* errcode_ret))dlsym(RTLD_NEXT, "clCreateCommandQueue");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueNDRangeKernel_p_ = (cl_int (*) (cl_command_queue command_queue,
        cl_kernel kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event))dlsym(RTLD_NEXT, "clEnqueueNDRangeKernel");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clSetKernelArg_p_ = (cl_int (*) (cl_kernel kernel,
                cl_uint arg_index,
                size_t arg_size,
                const void* arg_value))
        dlsym(RTLD_NEXT, "clSetKernelArg");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clCreateBuffer_p_ = (cl_mem (*) (cl_context context,
                cl_mem_flags flags,
                size_t size,
                void* host_ptr,
                cl_int* errcode_ret))
        dlsym(RTLD_NEXT, "clCreateBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueWriteBuffer_p_ = (cl_int CL_API_CALL (*) (
                cl_command_queue command_queue,
                cl_mem buffer,
                cl_bool blocking_write,
                size_t offset,
                size_t size,
                const void* ptr,
                cl_uint num_events_in_wait_list,
                const cl_event* event_wait_list,
                cl_event* event))dlsym(RTLD_NEXT, "clEnqueueWriteBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clGetEventInfo_p_ = (cl_int CL_API_CALL (*) (
                cl_event event,
                cl_profiling_info param_name,
                size_t param_value_size,
                void* param_value,
                size_t* param_value_size_ret))dlsym(RTLD_NEXT, "clGetEventInfo");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }
}
