//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) Computer Systems and Platforms Laboratory
// http://csap.snu.ac.kr
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLKernel.h"
#include <sys/time.h>
#include <sys/stat.h>
#include <algorithm>

static long long getTimestamp(void)
{
	static __thread struct timeval tv;
	gettimeofday(&tv, NULL);
	return 1000000LL * tv.tv_sec + tv.tv_usec;
}

ArmCLKernelProfiler::ArmCLKernelProfiler()
{
}

ArmCLKernelProfiler::~ArmCLKernelProfiler()
{
}

size_t* ArmCLKernelProfiler::get_reordered_map(size_t num_points)
{
    std::map<size_t, size_t*>::iterator it = reordered_map.find(num_points);

    if (it != reordered_map.end()) {
        size_t* map = it->second;
        return map;
    }

    return NULL;
}

void ArmCLKernelProfiler::reordering(size_t num_points, size_t* work_size)
{
    size_t* work_items = (size_t*)malloc(num_points * sizeof(size_t));
    for (int i = 0; i < num_points; i++)
        work_items[i] = i;

    std::sort(work_items, work_items + num_points,
            [&](size_t x, size_t y) {
                return work_size[x] < work_size[y];
            });

    reordered_map.insert(std::pair<size_t, size_t*>
            (num_points, work_items));

    //dump_reordered_map(num_points);
}

void ArmCLKernelProfiler::dump_reordered_map(size_t num_points)
{
    std::map<size_t, size_t*>::iterator it = reordered_map.find(num_points);

    if (it != reordered_map.end()) {
        size_t* map = it->second;

        FILE* file = fopen("dump.txt", "w");
        for (int i = 0; i < num_points; i++) {
            fprintf(file, "%d %zu\n", i, map[i]);
        }

        fclose(file);
    }
}

ArmCLKernel::ArmCLKernel(cl_kernel kernel, cl_kernel cpu_kernel,
        const char* kernel_name, ArmCLProgram* armcl_program)
    : kernel_(kernel), cpu_kernel_(cpu_kernel), kernel_name_(kernel_name),
    armcl_program_(armcl_program)
{
#ifdef DEBUG
    std::cout << "ArmCL: " << " kernel created" << std::endl;
#endif

    program_ = armcl_program_->program();
    cpu_program_ = armcl_program->cpu_program();
    armcl_context_ = armcl_program->armcl_context();
    context_ = armcl_context_->context();
    cpu_context_ = armcl_context_->cpu_context();
    //cpu_command_queue_ = armcl_context_->cpu_command_queue();

    PrepareOpenCLCalls();
    PreparePreprocessedKernels();
    PrepareKernelInformation();
    PrepareCPUThreadingKernels();
    PrepareContextInformation();

    // these arrays are for FinePar manual static workload partitioner
    finepar_remap_ = (int*)aligned_alloc(4096, sizeof(int) * max_work_items);
    for (int i = 0; i< max_work_items; i++)
        finepar_remap_[i] = i;
    is_finepar_partitioned_ = 0;

    kernel_profiler_ = new ArmCLKernelProfiler();
}

ArmCLKernel::~ArmCLKernel()
{
    free(finepar_remap_);

#ifdef DEBUG
    std::cout << "ArmCL: " << " kernel released" << std::endl;
#endif
}

#define FILE_DUMP
int ArmCLKernel::LoopProfilerRun(cl_command_queue command_queue,
        cl_kernel kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "ArmCLKernel: LoopProfilerRun" << std::endl;
    if (loop_profiler_kernel_ == NULL) {
        std::cout << "loop_profiler_kernel_ empty" << std::endl;
        return -1;
    }

    size_t total_work_size = 0;
    if (work_dim == 1) {
        total_work_size = global_work_size[0];
    } else if (work_dim == 2) {
        total_work_size = global_work_size[0] * global_work_size[1];
    } else if (work_dim == 3) {
        total_work_size = global_work_size[0] * global_work_size[1] *
            global_work_size[2];
    }

    std::cout << "total_work_size: " << total_work_size << std::endl;
    size_t size_loop_profile_data = total_work_size * sizeof(size_t);
    size_t* loop_profile_data = (size_t*)malloc(size_loop_profile_data);
    cl_int err;

    cl_mem m_loop_profile_data = clCreateBuffer_p_(context_,
            CL_MEM_READ_WRITE,
            size_loop_profile_data,
            NULL,
            &err);
    if (err != CL_SUCCESS) {
        std::cout << "clCreateBuffer failed for loop profile data"
            << std::endl;
        return -1;
    }

    err = clSetKernelArg_p_(loop_profiler_kernel_,
            loop_profiler_kernel_num_args_ - 1,
            sizeof(cl_mem),
            (void*)&m_loop_profile_data
            );
    if (err != CL_SUCCESS) {
        std::cout << "clSetKernelArg failed for loop profile data"
            << std::endl;
        return -1;
    }

    err = clEnqueueNDRangeKernel_p_(command_queue,
            //kernel_,
            loop_profiler_kernel_,
            work_dim,
            global_work_offset,
            global_work_size,
            local_work_size,
            num_events_in_wait_list,
            event_wait_list,
            event);
    if (err != CL_SUCCESS) {
        std::cout << "clEnqueueNDRangeKernel failed for loop profile data: "
            << " errcode: " << err
            << std::endl;
        return -1;
    }

    err = clEnqueueReadBuffer(command_queue,
            m_loop_profile_data,
            CL_TRUE,
            0,
            size_loop_profile_data,
            loop_profile_data,
            0,
            NULL,
            NULL);
    if (err != CL_SUCCESS) {
        std::cout << "clEnqueueReadBuffer failed for loop profile data"
            << std::endl;
        return -1;
    }

    kernel_profiler_->reordering(total_work_size, loop_profile_data);

#ifdef FILE_DUMP
    FILE* file_p = fopen("temp.txt", "w");
    if (file_p) {
        for (int i = 0; i < total_work_size; i++) {
            fprintf(file_p, "work item: %d work size: %zu\n",
                    i, loop_profile_data[i]);
        }
        fclose(file_p);
    }
#endif
    free(loop_profile_data);
}

cl_int ArmCLKernel::SetKernelArgSVMPointer(cl_uint arg_index, const void* arg_value)
{
    cl_int err;

#ifdef DEBUG
    std::cout << "ArmCLKernel: SetKernelArgSVMPointer" << std::endl;
#endif

    if (mbap_cpu_thread_kernel_) {
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            int index = i * mbap_cpu_thread_kernel_num_args_ + (int)arg_index;
            mbap_cpu_thread_kernel_params_[index] = (unsigned long)arg_value;
        }
    }

    if (mbap_cpu_thread_kernel_array_[0]) {
        for (int j = 0; j < 100; j++) {
            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                int index = i * mbap_cpu_thread_kernel_num_args_ + (int)arg_index;
                mbap_cpu_thread_kernel_array_params_[j][index] = (unsigned long)arg_value;
            }
        }
    }
}

cl_int ArmCLKernel::SetKernelArg(cl_uint arg_index, size_t arg_size,
        const void* arg_value, std::map<cl_mem, ArmCLBuffer*> armcl_buffers)
{
    cl_int err;

#ifdef DEBUG
    std::cout << "ArmCLKernel: SetKernelArg" << std::endl;
#endif

    ArmCLBuffer* armcl_buffer = NULL;
    cl_mem cpu_buffer = NULL;
    cl_mem* addr_ = (cl_mem*)arg_value;
    cl_mem_flags flags;
    size_t size;
    void* host_ptr = 0;

    for (std::map<cl_mem, ArmCLBuffer*>::iterator it =
            armcl_buffers.begin();
            it != armcl_buffers.end();
            it++) {
        cl_mem memobj = it->first;
        if (memobj == *addr_) {
#ifdef DEBUG
            std::cout << "ArmCLKernel: found the memobj at arg " << arg_index << std::endl;
#endif
            armcl_buffer = it->second;
            cpu_buffer = armcl_buffer->cpu_buffer();

            flags = armcl_buffer->flags();
            size = armcl_buffer->size();
            host_ptr = armcl_buffer->host_ptr();

            break;
        }
    }

    // MBAP technique
    if (mbap_profiler_kernel_) {
        err = clSetKernelArg_p_(mbap_profiler_kernel_,
                arg_index,
                arg_size,
                arg_value);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clSetKernelArg failed for mbap_profiler_kernel"
                << std::endl;
        }
    }

    if (mbap_kernel_) {
        err = clSetKernelArg_p_(mbap_kernel_,
                arg_index,
                arg_size,
                arg_value);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clSetKernelArg failed for mbap_kernel"
                << std::endl;
        }
    }

    if (mbap_cpu_thread_kernel_) {
        // CPU thread is enabled
        //printf("CPU thread is enabled\n");

        if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
            // this argument is opencl buffer (cl_mem)
            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                int index = i * mbap_cpu_thread_kernel_num_args_ + (int)arg_index;
                mbap_cpu_thread_kernel_params_[index] = (unsigned long)host_ptr;
            }
        } else if (armcl_buffer == NULL) {
            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                int index = i * mbap_cpu_thread_kernel_num_args_ + (int)arg_index;
                mbap_cpu_thread_kernel_params_[index] = (unsigned long)arg_value;
            }
        }
    }

    if (mbap_cpu_profiler_kernel_) {
        if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
            // this argument is opencl buffer (cl_mem)
            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                int index = i * mbap_cpu_profiler_kernel_num_args_ + (int)arg_index;
                mbap_cpu_profiler_kernel_params_[index] = (unsigned long)host_ptr;
            }
        } else if (armcl_buffer == NULL) {
            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                int index = i * mbap_cpu_profiler_kernel_num_args_ + (int)arg_index;
                mbap_cpu_profiler_kernel_params_[index] = (unsigned long)arg_value;
            }
        }
    }

    if (mbap_cpu_thread_kernel_array_[0]) {
        // CPU thread is enabled
        //printf("CPU thread is enabled\n");

        for (int j = 0; j < 100; j++) {
            if (armcl_buffer != NULL && flags & CL_MEM_USE_HOST_PTR) {
                // this argument is opencl buffer (cl_mem)
                for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                    int index = i * mbap_cpu_thread_kernel_num_args_ + (int)arg_index;
                    mbap_cpu_thread_kernel_array_params_[j][index] = (unsigned long)host_ptr;
                }
            } else if (armcl_buffer == NULL) {
                for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                    int index = i * mbap_cpu_thread_kernel_num_args_ + (int)arg_index;
                    mbap_cpu_thread_kernel_array_params_[j][index] = (unsigned long)arg_value;
                }
            }
        }
    }

    //
    if (cpu_kernel_) {
        if (cpu_buffer != NULL) {
            err = clSetKernelArg_p_(cpu_kernel_,
                    arg_index,
                    arg_size,
                    (void*)&cpu_buffer);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clSetKernelArg failed for cpu kernel"
                    << std::endl;
            }
        } else {
            err = clSetKernelArg_p_(cpu_kernel_,
                    arg_index,
                    arg_size,
                    arg_value);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clSetKernelArg failed for cpu kernel"
                    << std::endl;
            }
        }
    }

    if (loop_profiler_kernel_) {
        err = clSetKernelArg_p_(loop_profiler_kernel_,
                arg_index,
                arg_size,
                arg_value);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clSetKernelArg failed for loop profiler kernel"
                << std::endl;
        }
    }

    if (buf_profiler_kernel_) {
        err = clSetKernelArg_p_(buf_profiler_kernel_,
                arg_index,
                arg_size,
                arg_value);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clSetKernelArg failed for buf profiler kernel"
                << std::endl;
        }
    }

    if (cpu_threshold_run_kernel_) {
        if (cpu_buffer != NULL) {
            err = clSetKernelArg_p_(cpu_threshold_run_kernel_,
                    arg_index,
                    arg_size,
                    (void*)&cpu_buffer);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clSetKernelArg failed for cpu kernel"
                    << std::endl;
            }
        } else {
            err = clSetKernelArg_p_(cpu_threshold_run_kernel_,
                    arg_index,
                    arg_size,
                    arg_value);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clSetKernelArg failed for cpu kernel"
                    << std::endl;
            }
        }
    }

    if (gpu_threshold_run_kernel_) {
        err = clSetKernelArg_p_(gpu_threshold_run_kernel_,
                arg_index,
                arg_size,
                arg_value);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clSetKernelArg failed for gpu threshold run kernel"
                << std::endl;
        }
    }

    if (cpu_reordered_run_kernel_) {
        if (cpu_buffer != NULL) {
            err = clSetKernelArg_p_(cpu_reordered_run_kernel_,
                    arg_index,
                    arg_size,
                    (void*)&cpu_buffer);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clSetKernelArg failed for cpu reordered run kernel"
                    << std::endl;
            }
        } else {
            err = clSetKernelArg_p_(cpu_reordered_run_kernel_,
                    arg_index,
                    arg_size,
                    arg_value);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clSetKernelArg failed for cpu reordered run kernel"
                    << std::endl;
            }
        }
    }

    if (gpu_reordered_run_kernel_) {
        err = clSetKernelArg_p_(gpu_reordered_run_kernel_,
                arg_index,
                arg_size,
                arg_value);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clSetKernelArg failed for gpu reordered run kernel"
                << std::endl;
        }
    }

    return 0;
}

void ArmCLKernel::PrepareOpenCLCalls()
{
    char* error;

    clCreateKernel_p_ = (cl_kernel (*) (cl_program program,
                const char* kernel_name,
                cl_int* errcode_ret))dlsym(RTLD_NEXT, "clCreateKernel");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueNDRangeKernel_p_ = (cl_int (*) (cl_command_queue command_queue,
        cl_kernel kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event))dlsym(RTLD_NEXT, "clEnqueueNDRangeKernel");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clSetKernelArg_p_ = (cl_int (*) (cl_kernel kernel,
                cl_uint arg_index,
                size_t arg_size,
                const void* arg_value))
        dlsym(RTLD_NEXT, "clSetKernelArg");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clCreateBuffer_p_ = (cl_mem (*) (cl_context context,
                cl_mem_flags flags,
                size_t size,
                void* host_ptr,
                cl_int* errcode_ret))
        dlsym(RTLD_NEXT, "clCreateBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }

    clEnqueueWriteBuffer_p_ = (cl_int CL_API_CALL (*) (
                cl_command_queue command_queue,
                cl_mem buffer,
                cl_bool blocking_write,
                size_t offset,
                size_t size,
                const void* ptr,
                cl_uint num_events_in_wait_list,
                const cl_event* event_wait_list,
                cl_event* event))dlsym(RTLD_NEXT, "clEnqueueWriteBuffer");
    if ((error = dlerror()) != NULL) {
        fputs(error, stderr);
        exit(1);
    }
}

void ArmCLKernel::PreparePreprocessedKernels()
{
    std::string loop_profiler_kernel_name;
    std::string buf_profiler_kernel_name;
    std::string cpu_threshold_run_kernel_name;
    std::string gpu_threshold_run_kernel_name;
    std::string cpu_reordered_run_kernel_name;
    std::string gpu_reordered_run_kernel_name;

    /// MBAP technique
    std::string mbap_profiler_kernel_name;
    std::string mbap_kernel_name;

    loop_profiler_kernel_name.append(kernel_name_);
    loop_profiler_kernel_name.append("_loop_sampler");

    buf_profiler_kernel_name.append(kernel_name_);
    buf_profiler_kernel_name.append("_buf_sampler");

    cpu_threshold_run_kernel_name.append(kernel_name_);
    cpu_threshold_run_kernel_name.append("_cpu_threshold_run");

    gpu_threshold_run_kernel_name.append(kernel_name_);
    gpu_threshold_run_kernel_name.append("_gpu_threshold_run");

    cpu_reordered_run_kernel_name.append(kernel_name_);
    cpu_reordered_run_kernel_name.append("_cpu_reordered_run");

    gpu_reordered_run_kernel_name.append(kernel_name_);
    gpu_reordered_run_kernel_name.append("_gpu_reordered_run");

    mbap_profiler_kernel_name.append(kernel_name_);
    mbap_profiler_kernel_name.append("_profiler");

    mbap_kernel_name.append(kernel_name_);
    mbap_kernel_name.append("_OnTheFly");
    //mbap_kernel_name.append("_MBAP");

#ifdef DEBUG
    std::cout << "ArmCL: "
        << " prepare profiler run: "
        << " kernel_name: " << kernel_name_
        << " loop_profiler_kernel_name: " << loop_profiler_kernel_name
        << " buf_profiler_kernel_name: " << buf_profiler_kernel_name
        << " cpu_threshold_run_kernel_name: " << cpu_threshold_run_kernel_name
        << " gpu_threshold_run_kernel_name: " << gpu_threshold_run_kernel_name
        << " cpu_reordered_run_kernel_name: " << cpu_reordered_run_kernel_name
        << " gpu_reordered_run_kernel_name: " << gpu_reordered_run_kernel_name
        << " [MBAP] mbap_profiler_kernel_name: " << mbap_profiler_kernel_name
        << " [MBAP] mbap_kernel_name: " << mbap_kernel_name
        << std::endl;
#endif

    cl_int err;
    loop_profiler_kernel_ = clCreateKernel_p_(program_,
            loop_profiler_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare loop_profiler_kernel: "
            << " can't create loop_profiler_kernel "
            << loop_profiler_kernel_name
            << std::endl;
#endif
    }

    buf_profiler_kernel_ = clCreateKernel_p_(program_,
            buf_profiler_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare buf_profiler_kernel: "
            << " can't create buf_profiler_kernel "
            << buf_profiler_kernel_name
            << std::endl;
#endif
    }

    cpu_threshold_run_kernel_ = clCreateKernel_p_(cpu_program_,
            cpu_threshold_run_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare cpu_threshold_run_kernel: "
            << " can't create cpu_threshold_run_kernel "
            << cpu_threshold_run_kernel_name
            << std::endl;
#endif
    }

    gpu_threshold_run_kernel_ = clCreateKernel_p_(program_,
            gpu_threshold_run_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare gpu_threshold_run_kernel: "
            << " can't create gpu_threshold_run_kernel "
            << gpu_threshold_run_kernel_name
            << std::endl;
#endif
    }

    cpu_reordered_run_kernel_ = clCreateKernel_p_(cpu_program_,
            cpu_reordered_run_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare cpu_reordered_run_kernel: "
            << " can't create cpu_reordered_run_kernel "
            << cpu_reordered_run_kernel_name
            << std::endl;
#endif
    }

    gpu_reordered_run_kernel_ = clCreateKernel_p_(program_,
            gpu_reordered_run_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare gpu_reordered_run_kernel: "
            << " can't create gpu_reordered_run_kernel "
            << gpu_reordered_run_kernel_name
            << std::endl;
#endif
    }

    // prepare kernels for MBAP technique
    mbap_profiler_kernel_ = clCreateKernel_p_(program_,
            mbap_profiler_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare profiler_kernel: "
            << " can't create profiler_kernel "
            << mbap_profiler_kernel_name
            << std::endl;
#endif
    }

    mbap_kernel_ = clCreateKernel_p_(program_,
            mbap_kernel_name.c_str(), &err);
    if (err != CL_SUCCESS) {
#ifdef DEBUG
        std::cout << "ArmCL: "
            << " prepare mbap_kernel: "
            << " can't create mbap_kernel "
            << mbap_kernel_name
            << std::endl;
#endif
    }
}

// this is for MBAP technique
void ArmCLKernel::PrepareCPUThreadingKernels()
{
    std::string CPU_thread_kernel_file_name;
    CPU_thread_kernel_file_name.append(kernel_name_);
    CPU_thread_kernel_file_name.append("-CPU-prep.cpp");

    std::string CPU_thread_kernel_library_name;
    CPU_thread_kernel_library_name.append(".");
    CPU_thread_kernel_library_name.append(kernel_name_);
    CPU_thread_kernel_library_name.append("-CPU-prep.so");

    std::string command;
    command.append("g++ -std=c++11 -shared -fpic -o "); // use c++11 for std::atomic
    command.append(CPU_thread_kernel_library_name);
    command.append(" ");
    command.append(CPU_thread_kernel_file_name);
    struct stat buffer;
    if (stat (CPU_thread_kernel_file_name.c_str(), &buffer) == 0) {
        system(command.c_str()); // generate dynamic library of CPU kernel thread
    }
    //system(command.c_str()); // generate dynamic library of CPU kernel thread

    const char* current_path = std::getenv("PWD");
    std::string CPU_thread_kernel_path;
    CPU_thread_kernel_path.append(current_path);
    CPU_thread_kernel_path.append("/");
    CPU_thread_kernel_path.append(CPU_thread_kernel_library_name);
    void* executable = dlopen(CPU_thread_kernel_path.c_str(), RTLD_NOW);
    if (executable == NULL) {
#ifdef DEBUG
        fprintf(stderr, "executable is NULL\n");
#endif
    }

    *(void**)(&mbap_cpu_thread_kernel_) = dlsym(executable, "CPU_thread_entry");
    if (mbap_cpu_thread_kernel_ == NULL) {
#ifdef DEBUG
        fprintf(stderr, "mbap_cpu_thread_kernel_ is NULL\n");
#endif
    }

    mbap_cpu_thread_kernel_params_ = (unsigned long*)dlsym(executable,
            "x86_kernel_params");
    if (mbap_cpu_thread_kernel_params_) {
#ifdef DEBUG
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            for (int j = 0; j < mbap_cpu_thread_kernel_num_args_; j++) {
                fprintf(stdout, "%lu ", mbap_cpu_thread_kernel_params_[
                        i*mbap_cpu_thread_kernel_num_args_ + j]);
            }
            fprintf(stdout, "\n");
        }
#endif
    }

    std::string CPU_profiler_kernel_file_name;
    CPU_profiler_kernel_file_name.append(kernel_name_);
    CPU_profiler_kernel_file_name.append("-CPU-profiler-prep.cpp");

    std::string CPU_profiler_kernel_library_name;
    CPU_profiler_kernel_library_name.append(".");
    CPU_profiler_kernel_library_name.append(kernel_name_);
    CPU_profiler_kernel_library_name.append("-CPU-profiler-prep.so");

    std::string command_;
    command_.append("g++ -std=c++11 -shared -fpic -o ");
    command_.append(CPU_profiler_kernel_library_name);
    command_.append(" ");
    command_.append(CPU_profiler_kernel_file_name);
    struct stat buffer_;
    if (stat (CPU_profiler_kernel_file_name.c_str(), &buffer_) == 0) {
        system(command_.c_str()); // generate dynamic library of CPU kernel thread
    }
    //system(command_.c_str());

    const char* current_path_ = std::getenv("PWD");
    std::string CPU_profiler_kernel_path;
    CPU_profiler_kernel_path.append(current_path_);
    CPU_profiler_kernel_path.append("/");
    CPU_profiler_kernel_path.append(CPU_profiler_kernel_library_name);
    void* executable_ = dlopen(CPU_profiler_kernel_path.c_str(), RTLD_NOW);
    if (executable_ == NULL) {
#ifdef DEBUG
        fprintf(stderr, "executable is NULL\n");
#endif
    }

    *(void**)(&mbap_cpu_profiler_kernel_) = dlsym(executable_, "CPU_thread_entry");
    if (mbap_cpu_profiler_kernel_ == NULL) {
#ifdef DEBUG
        fprintf(stderr, "mbap_cpu_profiler_kernel_ is NULL\n");
#endif
    }

    mbap_cpu_profiler_kernel_params_ = (unsigned long*)dlsym(executable_,
            "x86_kernel_params");
    if (mbap_cpu_profiler_kernel_params_) {
#ifdef DEBUG
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            for (int j =0; j < mbap_cpu_profiler_kernel_num_args_; j++) {
                fprintf(stdout, "lu ", mbap_cpu_profiler_kernel_params_[
                        i*mbap_cpu_profiler_kernel_num_args_ + j]);
            }
            fprintf(stdout, "\n");
        }
#endif
    }

    mbap_cpu_thread_kernel_array_params_ = (unsigned long**)malloc(100 * sizeof(unsigned long*));
    for (int i = 0; i <100; i++) {
        std::string CPU_thread_kernel_file_name;
        CPU_thread_kernel_file_name.append(kernel_name_);
        CPU_thread_kernel_file_name.append("-CPU-prep.cpp");

        std::string CPU_thread_kernel_library_name;
        char buf[3];
        sprintf(buf, "%d", i);
        CPU_thread_kernel_library_name.append(".");
        CPU_thread_kernel_library_name.append(kernel_name_);
        CPU_thread_kernel_library_name.append(buf);
        CPU_thread_kernel_library_name.append("-CPU-prep.so");

        std::string command;
        command.append("g++ -std=c++11 -shared -fpic -o "); // use c++11 for std::atomic
        command.append(CPU_thread_kernel_library_name);
        command.append(" ");
        command.append(CPU_thread_kernel_file_name);
        struct stat buffer;
        if (stat (CPU_thread_kernel_file_name.c_str(), &buffer) == 0) {
            system(command.c_str()); // generate dynamic library of CPU kernel thread
        }
        //system(command.c_str()); // generate dynamic library of CPU kernel thread

        const char* current_path = std::getenv("PWD");
        std::string CPU_thread_kernel_path;
        CPU_thread_kernel_path.append(current_path);
        CPU_thread_kernel_path.append("/");
        CPU_thread_kernel_path.append(CPU_thread_kernel_library_name);
        void* executable = dlopen(CPU_thread_kernel_path.c_str(), RTLD_NOW);
        if (executable == NULL) {
#ifdef DEBUG
            fprintf(stderr, "executable is NULL\n");
#endif
        }

        *(void**)(&mbap_cpu_thread_kernel_array_[i]) =
            dlsym(executable, "CPU_thread_entry");
        if (mbap_cpu_thread_kernel_array_ == NULL) {
#ifdef DEBUG
            fprintf(stderr, "mbap_cpu_thread_kernel_ is NULL\n");
#endif
        }

        mbap_cpu_thread_kernel_array_params_[i] =
            (unsigned long*)dlsym(executable,
                    "x86_kernel_params");
        if (mbap_cpu_thread_kernel_array_params_[i]) {
#ifdef DEBUG
            for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
                for (int j = 0; j < mbap_cpu_thread_kernel_num_args_; j++) {
                    fprintf(stdout, "%lu ", mbap_cpu_thread_kernel_params_[
                            i*mbap_cpu_thread_kernel_num_args_ + j]);
                }
                fprintf(stdout, "\n");
            }
#endif
        }
    }

    // get information about proper core allocations
    int* num_proper_cpu_cores_p = (int*)dlsym(executable, "num_proper_cpu_cores");
    if (num_proper_cpu_cores_p != NULL)
        num_proper_cpu_cores_ = *num_proper_cpu_cores_p;
    else
        num_proper_cpu_cores_ = 0;

    int* num_proper_gpu_cores_p = (int*)dlsym(executable, "num_proper_gpu_cores");
    if (num_proper_gpu_cores_p != NULL)
        num_proper_gpu_cores_ = *num_proper_gpu_cores_p;
    else
        num_proper_gpu_cores_ = 0;

#ifdef DEBUG
    if (num_proper_cpu_cores_p != NULL && num_proper_gpu_cores_p != NULL)
        std::cout << kernel_name_ << " num_proper_cpu_cores: " << num_proper_cpu_cores_
            << " num_proper_gpu_cores: " << num_proper_gpu_cores_ << std::endl;
#endif

    char* num_proper_cpu_cores_c = std::getenv("NUM_CPU_CORES");
    if (num_proper_cpu_cores_c != NULL) {
        int num_proper_cpu_cores = atoi(num_proper_cpu_cores_c);
        num_proper_cpu_cores_ = num_proper_cpu_cores;
    }

    char* num_proper_gpu_cores_c = std::getenv("NUM_GPU_CORES");
    if (num_proper_gpu_cores_c != NULL) {
        int num_proper_gpu_cores = atoi(num_proper_gpu_cores_c);
        num_proper_gpu_cores_ = num_proper_gpu_cores;
    }

//#ifdef DEBUG
    std::cout << kernel_name_ << " num_proper_cpu_cores(ENV): " << num_proper_cpu_cores_
        << " num_proper_gpu_cores(ENV): " << num_proper_gpu_cores_ << std::endl;
//#endif
}

void ArmCLKernel::PrepareKernelInformation()
{
    clGetKernelInfo(kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &kernel_num_args_,
            NULL);

    clGetKernelInfo(loop_profiler_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &loop_profiler_kernel_num_args_,
            NULL);

    clGetKernelInfo(buf_profiler_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &buf_profiler_kernel_num_args_,
            NULL);

    clGetKernelInfo(cpu_threshold_run_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &cpu_threshold_run_kernel_num_args_,
            NULL);

    clGetKernelInfo(gpu_threshold_run_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &gpu_threshold_run_kernel_num_args_,
            NULL);

    clGetKernelInfo(cpu_reordered_run_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &cpu_reordered_run_kernel_num_args_,
            NULL);

    clGetKernelInfo(gpu_reordered_run_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &gpu_reordered_run_kernel_num_args_,
            NULL);

    // MBAP technique
    clGetKernelInfo(mbap_profiler_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &mbap_profiler_kernel_num_args_,
            NULL);

    clGetKernelInfo(mbap_kernel_,
            CL_KERNEL_NUM_ARGS,
            sizeof(cl_uint),
            &mbap_kernel_num_args_,
            NULL);

    mbap_cpu_thread_kernel_num_args_ =
        (int)kernel_num_args_ + 5;

    mbap_cpu_profiler_kernel_num_args_ =
        (int)kernel_num_args_ + 5;

#ifdef DEBUG
    std::cout << "kernel: " << kernel_name_
        << " num_args: " << kernel_num_args_
        << std::endl;

    std::cout << "loop_profiler_kernel: "
        << " num_args: " << loop_profiler_kernel_num_args_
        << std::endl;

    std::cout << "buf_profiler_kernel: "
        << " num_args: " << buf_profiler_kernel_num_args_
        << std::endl;

    std::cout << "cpu_threshold_run_kernel: "
        << " num_args: " << cpu_threshold_run_kernel_num_args_
        << std::endl;

    std::cout << "gpu_threshold_run_kernel: "
        << " num_args: " << gpu_threshold_run_kernel_num_args_
        << std::endl;

    std::cout << "cpu_reordered_run_kernel: "
        << " num_args: " << cpu_reordered_run_kernel_num_args_
        << std::endl;

    std::cout << "gpu_reordered_run_kernel: "
        << " num_args: " << gpu_reordered_run_kernel_num_args_
        << std::endl;

    std::cout << "[MBAP] profiler_kernel: "
        << " num_args: " << mbap_profiler_kernel_num_args_
        << std::endl;

    std::cout << "[MBAP] mbap_kernel: "
        << " num_args: " << mbap_kernel_num_args_
        << std::endl;

    std::cout << "[MBAP] mbap_cpu_thread_kernel: "
        << " num_args: " << mbap_cpu_thread_kernel_num_args_
        << std::endl;
#endif
}

void ArmCLKernel::PrepareContextInformation()
{
    cl_uint err = clGetProgramInfo(program_,
            CL_PROGRAM_CONTEXT,
            sizeof(cl_context),
            &context_,
            NULL);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL: " << "clGetProgramInfo: failed" << std::endl;
    }
}
