//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) Computer Systems and Platforms Laboratory
// http://csap.snu.ac.kr
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#ifndef __ARM_OpenCL_Kernel_H_
#define __ARM_OpenCL_Kernel_H_

#include <CL/cl.h>
#include <iostream>
#include <dlfcn.h>
#include <stdio.h>
#include <map>
#include <vector>
#include <list>

#include "ArmCLContext.h"
#include "ArmCLProgram.h"
#include "ArmCLBuffer.h"

class ArmCLKernelProfiler {
  public:
    ArmCLKernelProfiler();
    ~ArmCLKernelProfiler();

    size_t* get_reordered_map(size_t num_points);

    void reordering(size_t num_points, size_t* work_size);

    void dump_reordered_map(size_t num_points);

  private:
    std::map<size_t, size_t*> reordered_map;
};

class ArmCLKernel {
  public:
    ArmCLKernel(cl_kernel kernel, cl_kernel cpu_kernel,
            const char* kernel_name, ArmCLProgram* armcl_program);
    ~ArmCLKernel();

    int LoopProfilerRun(cl_command_queue command_queue,
            cl_kernel kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    const char* kernel_name() { return kernel_name_; }

    cl_kernel kernel() const {
        return kernel_;
    }
    cl_kernel cpu_kernel() const {
        return cpu_kernel_;
    }

    // MBAP kernels
    cl_kernel mbap_profiler_kernel() const {
        return mbap_profiler_kernel_;
    }
    cl_kernel mbap_kernel() const {
        return mbap_kernel_;
    }
    int (*mbap_cpu_thread_kernel())(int, void*) const {
        return mbap_cpu_thread_kernel_;
    }
    int (*mbap_cpu_profiler_kernel())(int, void*) const {
        return mbap_cpu_profiler_kernel_;
    }

    cl_uint mbap_profiler_kernel_num_args() const {
        return mbap_profiler_kernel_num_args_;
    }
    cl_uint mbap_kernel_num_args() const {
        return mbap_kernel_num_args_;
    }
    int mbap_cpu_thread_kernel_num_args() const {
        return mbap_cpu_thread_kernel_num_args_;
    }
    int mbap_cpu_profiler_kernel_num_args() const {
        return mbap_cpu_profiler_kernel_num_args_;
    }

    unsigned long* mbap_cpu_thread_kernel_params() const {
        return mbap_cpu_thread_kernel_params_;
    }
    unsigned long* mbap_cpu_profiler_kernel_params() const {
        return mbap_cpu_profiler_kernel_params_;
    }

    int (*mbap_cpu_thread_kernel_array(int id))(int, void*) const {
        return mbap_cpu_thread_kernel_array_[id];
    }

    unsigned long* mbap_cpu_thread_kernel_array_params(int id) const {
        return mbap_cpu_thread_kernel_array_params_[id];
    }

    // proper core counts from MBAP technique
    int num_proper_cpu_cores() const {
        return num_proper_cpu_cores_;
    }
    int num_proper_gpu_cores() const {
        return num_proper_gpu_cores_;
    }

    // 
    cl_kernel loop_profiler_kernel() const {
        return loop_profiler_kernel_;
    }
    cl_kernel buf_profiler_kernel() const {
        return buf_profiler_kernel_;
    }
    cl_kernel cpu_threshold_run_kernel() const {
        return cpu_threshold_run_kernel_;
    }
    cl_kernel gpu_threshold_run_kernel() const {
        return gpu_threshold_run_kernel_;
    }
    cl_kernel cpu_reordered_run_kernel() const {
        return cpu_reordered_run_kernel_;
    }
    cl_kernel gpu_reordered_run_kernel() const {
        return gpu_reordered_run_kernel_;
    }

    cl_uint loop_profiler_kernel_num_args() const {
        return loop_profiler_kernel_num_args_;
    }
    cl_uint buf_profiler_kernel_num_args() const {
        return buf_profiler_kernel_num_args_;
    }
    cl_uint cpu_threshold_run_kernel_num_args() const {
        return cpu_threshold_run_kernel_num_args_;
    }
    cl_uint gpu_threshold_run_kernel_num_args() const {
        return gpu_threshold_run_kernel_num_args_;
    }
    cl_uint cpu_reordered_run_kernel_num_args() const {
        return cpu_reordered_run_kernel_num_args_;
    }
    cl_uint gpu_reordered_run_kernel_num_args() const {
        return gpu_reordered_run_kernel_num_args_;
    }

    // this is for FinePar manual static workload partitioner
    // assuming that this kernel is always executed with the same number of work-items
    int is_finepar_partitioned_;
    int finepar_threshold_;
    int* finepar_remap_;
    int finepar_cpu_wgs_;
    int finepar_gpu_wgs_;
    const int max_work_items = 10000000;

    ArmCLKernelProfiler* kernel_profiler() const {
        return kernel_profiler_;
    }

    cl_int SetKernelArgSVMPointer(cl_uint arg_index, const void* arg_value);

    cl_int SetKernelArg(cl_uint arg_index, size_t arg_size,
            const void* arg_value,
            std::map<cl_mem, ArmCLBuffer*> armcl_buffers);

  private:
    void PrepareOpenCLCalls();
    void PreparePreprocessedKernels();
    void PrepareCPUThreadingKernels();
    void PrepareKernelInformation();
    void PrepareContextInformation();

    const cl_kernel kernel_;
    const cl_kernel cpu_kernel_;
    const char* kernel_name_;
    cl_program program_;
    cl_program cpu_program_;
    cl_context context_;
    cl_context cpu_context_;
    cl_command_queue cpu_command_queue_;

    ArmCLProgram* armcl_program_;
    ArmCLContext* armcl_context_;

    cl_uint kernel_num_args_;
    cl_uint loop_profiler_kernel_num_args_;
    cl_uint buf_profiler_kernel_num_args_;
    cl_uint cpu_threshold_run_kernel_num_args_;
    cl_uint gpu_threshold_run_kernel_num_args_;
    cl_uint cpu_reordered_run_kernel_num_args_;
    cl_uint gpu_reordered_run_kernel_num_args_;

    cl_kernel loop_profiler_kernel_;
    cl_kernel buf_profiler_kernel_;
    cl_kernel cpu_threshold_run_kernel_;
    cl_kernel gpu_threshold_run_kernel_;
    cl_kernel cpu_reordered_run_kernel_;
    cl_kernel gpu_reordered_run_kernel_;

    /// code generation of current MBAP technique
    cl_uint mbap_profiler_kernel_num_args_;
    cl_uint mbap_kernel_num_args_;
    cl_kernel mbap_profiler_kernel_;
    cl_kernel mbap_kernel_;
    int mbap_cpu_thread_kernel_num_args_;
    int (*mbap_cpu_thread_kernel_)(int id, void* argp);
    unsigned long* mbap_cpu_thread_kernel_params_;
    int mbap_cpu_profiler_kernel_num_args_;
    int (*mbap_cpu_profiler_kernel_)(int id, void* argp);
    unsigned long* mbap_cpu_profiler_kernel_params_;

    int (*mbap_cpu_thread_kernel_array_[100])(int id, void* argp);
    unsigned long** mbap_cpu_thread_kernel_array_params_;

    // proper core counts from MBAP technique
    int num_proper_cpu_cores_;
    int num_proper_gpu_cores_;

    ArmCLKernelProfiler* kernel_profiler_;

    cl_kernel CL_API_CALL (*clCreateKernel_p_) (cl_program program,
            const char* kernel_name,
            cl_int* errcode_ret);

    cl_int CL_API_CALL (*clEnqueueNDRangeKernel_p_) (
            cl_command_queue command_queue,
            cl_kernel kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);

    cl_int CL_API_CALL (*clSetKernelArg_p_)(cl_kernel kernel,
            cl_uint arg_index,
            size_t arg_size,
            const void* arg_value);

    cl_mem CL_API_CALL (*clCreateBuffer_p_)(cl_context context,
            cl_mem_flags flags,
            size_t size,
            void* host_ptr,
            cl_int* errcode_ret);

    cl_int CL_API_CALL (*clEnqueueWriteBuffer_p_)(
            cl_command_queue command_queue,
            cl_mem buffer,
            cl_bool blocking_write,
            size_t offset,
            size_t size,
            const void* ptr,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);
};

#endif
