//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) 2017 Computer Systems and Platforms Laboratory.
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "armcl.h"

CL_API_ENTRY cl_context CL_API_CALL clCreateContext(
        const cl_context_properties* properties,
        cl_uint num_devices,
        const cl_device_id* devices,
        void (*pfn_notify) (
            const char* effinfo,
            const void* private_info,
            size_t cb,
            void* user_data
            ),
        void* user_data,
        cl_int *errcode_ret)
{
    static cl_context (*clCreateContext_p) (
            const cl_context_properties* properties,
            cl_uint num_devices,
            const cl_device_id* devices,
            void (*pfn_notify) (
                const char* effinfo,
                const void* private_info,
                size_t cb,
                void* user_data
                ),
            void* user_data,
            cl_int *errcode_ret);
    char* error;

#ifdef DEBUG
    std::cout << "clCreateContext call" << std::endl;
#endif

    if (!clCreateContext_p) {
        clCreateContext_p = (cl_context (*) (
                    const cl_context_properties* properties,
                    cl_uint num_devices,
                    const cl_device_id* devices,
                    void (*pfn_notify) (
                        const char* effinfo,
                        const void* private_info,
                        size_t cb,
                        void* user_data
                        ),
                    void* user_data,
                    cl_int *errcode_ret))dlsym(RTLD_NEXT, "clCreateContext");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    cl_context context = clCreateContext_p(properties,
            num_devices,
            devices,
            pfn_notify,
            user_data,
            errcode_ret);

    ArmCLContext* armcl_context = new ArmCLContext(context);
    armcl_contexts.insert(std::pair<cl_context, ArmCLContext*>
            (context, armcl_context));

    return context;
}

cl_int clReleaseContext(cl_context context)
{
    static cl_int (*clReleaseContext_p) (cl_context context);
    char* error;

#ifdef DEBUG
    std::cout << "clReleaseContext call" << std::endl;
#endif

    if (!clReleaseContext_p) {
        clReleaseContext_p = (cl_int (*) (cl_context context))
            dlsym(RTLD_NEXT, "clReleaseContext");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_context, ArmCLContext*>::iterator it =
        armcl_contexts.find(context);
    if (it != armcl_contexts.end()) {
        ArmCLContext* armcl_context = it->second;
        armcl_contexts.erase(context);
        delete armcl_context;
    } else {
        std::cout << "ArmCL: "
            << "unexpected: can't find the context"
            << std::endl;
    }

    return clReleaseContext_p(context);
}

CL_API_ENTRY cl_command_queue CL_API_CALL
clCreateCommandQueue(cl_context context,
        cl_device_id device,
        cl_command_queue_properties properties,
        cl_int* errcode_ret)
{
    static cl_command_queue CL_API_CALL (*clCreateCommandQueue_p)
        (cl_context context,
         cl_device_id device,
         cl_command_queue_properties properties,
         cl_int* errcode_ret);
    char* error;

#ifdef DEBUG
    std::cout << "clCreateCommandQueue call" << std::endl;
#endif

    if (!clCreateCommandQueue_p) {
        clCreateCommandQueue_p = (cl_command_queue CL_API_CALL (*)
                (cl_context context,
                 cl_device_id device,
                 cl_command_queue_properties properties,
                 cl_int* errcode_ret)
                )dlsym(RTLD_NEXT, "clCreateCommandQueue");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_context, ArmCLContext*>::iterator it =
        armcl_contexts.find(context);
    if (it != armcl_contexts.end()) {
        ArmCLContext* armcl_context = it->second;

        cl_command_queue command_queue = clCreateCommandQueue_p(
                context,
                device,
                properties,
                errcode_ret);

        ArmCLCommandQueue* armcl_command_queue = new ArmCLCommandQueue(
                command_queue,
                armcl_context);

        armcl_command_queues.insert(
                std::pair<cl_command_queue, ArmCLCommandQueue*>(
                    command_queue, armcl_command_queue));

        return command_queue;
    } else {
        std::cout << "ArmCL: "
            << "unexpected: can't find the context"
            << std::endl;
    }

    return clCreateCommandQueue_p(context,
            device,
            properties,
            errcode_ret);
}

CL_API_ENTRY cl_program CL_API_CALL
clCreateProgramWithSource(cl_context context,
        cl_uint count,
        const char ** strings,
        const size_t * lengths,
        cl_int * errcode_ret)
{
    static cl_program CL_API_CALL (*clCreateProgramWithSource_p)
        (cl_context context,
         cl_uint count,
         const char ** strings,
         const size_t * lengths,
         cl_int * errcode_ret);
    char* error;

#ifdef DEBUG
    std::cout << "clCreateProgramWithSource call" << std::endl;
#endif

    if (!clCreateProgramWithSource_p) {
        clCreateProgramWithSource_p = (cl_program CL_API_CALL (*)
                (cl_context context,
                 cl_uint count,
                 const char ** strings,
                 const size_t * lengths,
                 cl_int * errcode_ret)
                )dlsym(RTLD_NEXT, "clCreateProgramWithSource");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_context, ArmCLContext*>::iterator it =
        armcl_contexts.find(context);
    if (it != armcl_contexts.end()) {
        ArmCLContext* armcl_context = it->second;

        char* generated_source_str = (char*)malloc(0x100000);
        size_t generated_source_size;

        armcl_context->PreprocessingWithSource(count, strings, lengths);
        FILE* fp = fopen("./temp-prep.cl", "r");
        if (!fp) {
            fprintf(stderr, "Failed to load kernel.\n");
            exit(-1);
        }

        generated_source_size = fread(generated_source_str, 1, 0x100000, fp);
        fclose(fp);

        std::cout << "Program Building" << std::endl;

        cl_program program = clCreateProgramWithSource_p(context,
                count,
                (const char**)&generated_source_str,
                (const size_t*)&generated_source_size,
                //strings,
                //lengths,
                errcode_ret);

        cl_context cpu_context = armcl_context->cpu_context();

        cl_int err;
        cl_program cpu_program = clCreateProgramWithSource_p(cpu_context,
                count,
                (const char**)&generated_source_str,
                (const size_t*)&generated_source_size,
                //strings,
                //lengths,
                &err);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clCreateProgramWithSource failed for "
                << " cpu_program" << std::endl;
        }

        ArmCLProgram* armcl_program = new ArmCLProgram(program, cpu_program,
                armcl_context);
        //ArmCLProgram* armcl_program = new ArmCLProgram(program, context);
        armcl_programs.insert(std::pair<cl_program, ArmCLProgram*>
                (program, armcl_program));
        armcl_context->RegisterPrograms(armcl_program);

        return program;
    } else {
        std::cout << "ArmCL: "
            << "unexpected: can't find the context"
            << std::endl;
    }

    return clCreateProgramWithSource_p(context,
            count,
            strings,
            lengths,
            errcode_ret);
}

cl_int CL_API_CALL clBuildProgram(cl_program program,
        cl_uint num_devices,
        const cl_device_id* device_list,
        const char* options,
        void (CL_CALLBACK * pfn_notify)(cl_program program, void* user_data),
        void* user_data)
{
    static cl_int CL_API_CALL (*clBuildProgram_p) (cl_program program,
        cl_uint num_devices,
        const cl_device_id* device_list,
        const char* options,
        void (CL_CALLBACK * pfn_notify)(cl_program program, void* user_data),
        void* user_data);
    char* error;

#ifdef DEBUG
    std::cout << "clBuildProgram call" << std::endl;
#endif

    if (!clBuildProgram_p) {
        clBuildProgram_p = (cl_int CL_API_CALL (*) (cl_program program,
                    cl_uint num_devices,
                    const cl_device_id* device_list,
                    const char* options,
                    void (CL_CALLBACK * pfn_notify)(cl_program program,
                        void* user_data),
                    void* user_data))dlsym(RTLD_NEXT, "clBuildProgram");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_program, ArmCLProgram*>::iterator it =
        armcl_programs.find(program);
    if (it != armcl_programs.end()) {
        ArmCLProgram* armcl_program = it->second;
        armcl_program->BuildCPUProgram();
    }

    return clBuildProgram_p(program,
            num_devices,
            device_list,
            options,
            pfn_notify,
            user_data);
}

cl_int CL_API_CALL clReleaseProgram(cl_program program)
{
    static cl_int CL_API_CALL (*clReleaseProgram_p) (cl_program program);
    char* error;

#ifdef DEBUG
    std::cout << "clReleaseProgram call" << std::endl;
#endif

    if (!clReleaseProgram_p) {
        clReleaseProgram_p = (cl_int CL_API_CALL (*) (cl_program program))
            dlsym(RTLD_NEXT, "clReleaseProgram");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_program, ArmCLProgram*>::iterator it =
        armcl_programs.find(program);
    if (it != armcl_programs.end()) {
        ArmCLProgram* armcl_program = it->second;
        armcl_programs.erase(program);
        delete armcl_program;
    } else {
        std::cout << "ArmCL: "
            << "unexpected: can't find the context"
            << std::endl;
    }

    return clReleaseProgram_p(program);
}

cl_mem CL_API_CALL clCreateBuffer(cl_context context,
        cl_mem_flags flags,
        size_t size,
        void* host_ptr,
        cl_int* errcode_ret)
{
    static cl_mem CL_API_CALL (*clCreateBuffer_p) (cl_context context,
            cl_mem_flags flags,
            size_t size,
            void* host_ptr,
            cl_int* errcode_ret);
    char* error;

#ifdef DEBUG
    std::cout << "clCreateBuffer call" << std::endl;
#endif

    if (!clCreateBuffer_p) {
        clCreateBuffer_p = (cl_mem (*) (cl_context context,
                    cl_mem_flags flags,
                    size_t size,
                    void* host_ptr,
                    cl_int* errcode_ret))
            dlsym(RTLD_NEXT, "clCreateBuffer");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    cl_mem buffer = clCreateBuffer_p(context,
            flags,
            size,
            host_ptr,
            errcode_ret);

    std::map<cl_context, ArmCLContext*>::iterator it =
        armcl_contexts.find(context);
    if (it != armcl_contexts.end()) {
        ArmCLContext* armcl_context = it->second;
        cl_context cpu_context = armcl_context->cpu_context();

        cl_int err;

        cl_mem cpu_buffer = clCreateBuffer_p(cpu_context,
                flags,
                size,
                host_ptr,
                &err);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clCreateBuffer_p failed for cpu buffer"
                << std::endl;
        }

        ArmCLBuffer* armcl_buffer = new ArmCLBuffer(buffer,
                flags,
                size,
                host_ptr,
                cpu_buffer,
                armcl_context);

        //std::cout << "insert: " << buffer << std::endl;
        armcl_buffers.insert(std::pair<cl_mem, ArmCLBuffer*>
                (buffer, armcl_buffer));
    }

    return buffer;
}

cl_kernel CL_API_CALL clCreateKernel(cl_program program,
        const char* kernel_name,
        cl_int* errcode_ret)
{
    static cl_kernel (*clCreateKernel_p) (cl_program program,
            const char* kernel_name,
            cl_int* errcode_ret);
    char* error;

#ifdef DEBUG
    std::cout << "clCreateKernel call" << std::endl;
#endif

    if (!clCreateKernel_p) {
        clCreateKernel_p = (cl_kernel (*) (cl_program program,
                    const char* kernel_name,
                    cl_int* errcode_ret))dlsym(RTLD_NEXT, "clCreateKernel");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_program, ArmCLProgram*>::iterator it =
        armcl_programs.find(program);
    if (it != armcl_programs.end()) {
        ArmCLProgram* armcl_program = it->second;

        cl_kernel kernel = clCreateKernel_p(program,
                kernel_name,
                errcode_ret);

        cl_program cpu_program = armcl_program->cpu_program();

        //char param_src_value[40960];
        //fprintf(stdout, "output: %s\n", param_src_value);

        //cl_uint err = clGetProgramInfo(cpu_program,
        //        CL_PROGRAM_SOURCE,
        //        40960,
        //        param_src_value,
        //        NULL);

        //fprintf(stdout, "output: %s\n", param_src_value);

        //std::cout << "kernel_name: " << kernel_name << std::endl;
        cl_int err;
        cl_kernel cpu_kernel = clCreateKernel_p(cpu_program,
                kernel_name,
                &err);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: "
                << " clCreateKernel failed for cpu kernel" << std::endl;
        }
        //std::cout << "err: " << error << std::endl;

        ArmCLKernel* armcl_kernel = new ArmCLKernel(kernel, cpu_kernel,
                kernel_name,
                armcl_program);

        armcl_kernels.insert(std::pair<cl_kernel, ArmCLKernel*>
                (kernel, armcl_kernel));

        armcl_program->RegisterKernels(armcl_kernel);

        return kernel;
    }

    return clCreateKernel_p(program,
            kernel_name,
            errcode_ret);
}

//void* CL_API_CALL clSVMAlloc(cl_context context,
//        cl_svm_mem_flag flags,
//        size_t size,
//        unsigned int alignment)
//{
//    static void* CL_API_CALL (*clSVMAlloc_p) (cl_context context,
//            cl_svm_mem_flag flags,
//            size_t size,
//            unsigned int alignment);
//    char* error;
//}

cl_int CL_API_CALL clSetKernelArgSVMPointer(cl_kernel kernel,
        cl_uint arg_index,
        const void* arg_value)
{
    static cl_int CL_API_CALL (*clSetKernelArgSVMPointer_p) (cl_kernel kernel,
            cl_uint arg_index,
            const void* arg_value);
    char* error;

    if (!clSetKernelArgSVMPointer_p) {
        clSetKernelArgSVMPointer_p = (cl_int (*) (cl_kernel kernel,
                    cl_uint arg_index,
                    const void* arg_value))
            dlsym(RTLD_NEXT, "clSetKernelArg");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_kernel, ArmCLKernel*>::iterator it =
        armcl_kernels.find(kernel);
    if (it != armcl_kernels.end()) {
        ArmCLKernel* armcl_kernel = it->second;
        armcl_kernel->SetKernelArgSVMPointer(arg_index, arg_value);
    }

    return clSetKernelArgSVMPointer_p(kernel,
            arg_index,
            arg_value);
}

cl_int CL_API_CALL clSetKernelArg(cl_kernel kernel,
        cl_uint arg_index,
        size_t arg_size,
        const void* arg_value)
{
    static cl_int CL_API_CALL (*clSetKernelArg_p) (cl_kernel kernel,
            cl_uint arg_index,
            size_t arg_size,
            const void* arg_value);
    char* error;

#ifdef DEBUG
    std::cout << "clSetKernelArg call" << std::endl;
#endif

    if (!clSetKernelArg_p) {
        clSetKernelArg_p = (cl_int (*) (cl_kernel kernel,
                    cl_uint arg_index,
                    size_t arg_size,
                    const void* arg_value))
            dlsym(RTLD_NEXT, "clSetKernelArg");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_kernel, ArmCLKernel*>::iterator it=
        armcl_kernels.find(kernel);
    if (it != armcl_kernels.end()) {
        ArmCLKernel* armcl_kernel = it->second;
        armcl_kernel->SetKernelArg(arg_index, arg_size, arg_value,
                armcl_buffers);
    }

    return clSetKernelArg_p(kernel,
            arg_index,
            arg_size,
            arg_value);
}

//// This API is not called in HSA system if buffers are allocated
//// with flag of CL_MEM_USE_HOST_PTR
cl_int CL_API_CALL clEnqueueWriteBuffer(cl_command_queue command_queue,
        cl_mem buffer,
        cl_bool blocking_write,
        size_t offset,
        size_t size,
        const void* ptr,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    static cl_int CL_API_CALL (*clEnqueueWriteBuffer_p) (
            cl_command_queue command_queue,
            cl_mem buffer,
            cl_bool blocking_write,
            size_t offset,
            size_t size,
            const void* ptr,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);
    char* error;

#ifdef DEBUG
    std::cout << "clEnqueueWriteBuffer call" << std::endl;
#endif

    if (!clEnqueueWriteBuffer_p) {
        clEnqueueWriteBuffer_p = (cl_int CL_API_CALL (*) (
                    cl_command_queue command_queue,
                    cl_mem buffer,
                    cl_bool blocking_write,
                    size_t offset,
                    size_t size,
                    const void* ptr,
                    cl_uint num_events_in_wait_list,
                    const cl_event* event_wait_list,
                    cl_event* event))
            dlsym(RTLD_NEXT, "clEnqueueWriteBuffer");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    cl_int errcode_ret = clEnqueueWriteBuffer_p(command_queue,
            buffer,
            blocking_write,
            offset,
            size,
            ptr,
            num_events_in_wait_list,
            event_wait_list,
            event);

    std::map<cl_mem, ArmCLBuffer*>::iterator it =
        armcl_buffers.find(buffer);
    if (it != armcl_buffers.end()) {
        ArmCLBuffer* armcl_buffer = it->second;
        cl_mem cpu_buffer = armcl_buffer->cpu_buffer();

        std::map<cl_command_queue, ArmCLCommandQueue*>::iterator cq_it =
            armcl_command_queues.find(command_queue);
        if (cq_it != armcl_command_queues.end()) {
            ArmCLCommandQueue* armcl_command_queue = cq_it->second;
            cl_command_queue cpu_command_queue =
                armcl_command_queue->cpu_command_queue();

            cl_int err;
            err = clEnqueueWriteBuffer_p(cpu_command_queue,
                    cpu_buffer,
                    blocking_write,
                    offset,
                    size,
                    ptr,
                    num_events_in_wait_list,
                    event_wait_list,
                    event);
            if (err != CL_SUCCESS) {
                std::cout << "ArmCL: "
                    << " clEnqueueWriteBuffer failed for cpu buffer"
                    << std::endl;
            }
        }

        //ArmCLContext* armcl_context = armcl_buffer->armcl_context();
        //cl_command_queue cpu_command_queue = armcl_context->cpu_command_queue();

        //cl_int err;
        //err = clEnqueueWriteBuffer_p(cpu_command_queue,
        //        cpu_buffer,
        //        blocking_write,
        //        offset,
        //        size,
        //        ptr,
        //        num_events_in_wait_list,
        //        event_wait_list,
        //        event);
        //if (err != CL_SUCCESS) {
        //    std::cout << "ArmCL: "
        //        << " clEnqueueWriteBuffer failed for cpu buffer" << std::endl;
        //}
    }

    return errcode_ret;
}

cl_int CL_API_CALL clReleaseKernel(cl_kernel kernel)
{
    static cl_int (*clReleaseKernel_p) (cl_kernel kernel);
    char* error;

#ifdef DEBUG
    std::cout << "clReleaseKernel call" << std::endl;
#endif

    if (!clReleaseKernel_p) {
        clReleaseKernel_p = (cl_int (*) (cl_kernel kernel))
            dlsym(RTLD_NEXT, "clReleaseKernel");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

    std::map<cl_kernel, ArmCLKernel*>::iterator it =
        armcl_kernels.find(kernel);
    if (it != armcl_kernels.end()) {
        ArmCLKernel* armcl_kernel = it->second;
        armcl_kernels.erase(kernel);
        delete armcl_kernel;
    }

    return clReleaseKernel_p(kernel);
}

cl_int clEnqueueNDRangeKernel(cl_command_queue command_queue,
        cl_kernel kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    static cl_int (*clEnqueueNDRangeKernel_p) (cl_command_queue command_queue,
            cl_kernel kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event);
    char* error;

#ifdef DEBUG
    std::cout << "clEnqueueNDRangeKernel call" << std::endl;
#endif

    if (!clEnqueueNDRangeKernel_p) {
            clEnqueueNDRangeKernel_p = (cl_int (*) (cl_command_queue command_queue,
            cl_kernel kernel,
            cl_uint work_dim,
            const size_t* global_work_offset,
            const size_t* global_work_size,
            const size_t* local_work_size,
            cl_uint num_events_in_wait_list,
            const cl_event* event_wait_list,
            cl_event* event))dlsym(RTLD_NEXT, "clEnqueueNDRangeKernel");
        if ((error = dlerror()) != NULL) {
            fputs(error, stderr);
            exit(1);
        }
    }

#ifdef DEBUG
    std::cout << "global_work_size[0]: " << global_work_size[0]
        << " global_work_size[1]: " << global_work_size[1]
        << " global_work_size[2]: " << global_work_size[2] << std::endl;
#endif

    std::map<cl_kernel, ArmCLKernel*>::iterator kernel_it =
        armcl_kernels.find(kernel);
    ArmCLKernel* armcl_kernel = kernel_it->second;

    std::map<cl_command_queue, ArmCLCommandQueue*>::iterator cq_it =
        armcl_command_queues.find(command_queue);
    ArmCLCommandQueue* armcl_command_queue = cq_it->second;

    char option[80] = { 0, };
    strncpy(option, std::getenv("METHOD"), strlen(std::getenv("METHOD")));

    cl_kernel mbap_profiler_kernel = armcl_kernel->mbap_profiler_kernel();
    if (mbap_profiler_kernel != NULL && option != NULL) {
        printf("option: %s\n", option);

        if (!strncmp("CPU", option, strlen(option))) {
            printf("this is CPU option\n");

            return armcl_command_queue->CPUThreading(armcl_kernel,
                    work_dim,
                    global_work_offset,
                    global_work_size,
                    local_work_size,
                    num_events_in_wait_list,
                    event_wait_list,
                    event);
        } else if (!strncmp("GPU", option, strlen(option))) {
            printf("this is GPU option\n");

            return clEnqueueNDRangeKernel_p(command_queue,
                    kernel,
                    work_dim,
                    global_work_offset,
                    global_work_size,
                    local_work_size,
                    num_events_in_wait_list,
                    event_wait_list,
                    event);
        } else if (!strncmp("WorkGroupPartitioning", option, strlen(option))) {
            printf("Workgroup partitioning\n");

            return armcl_command_queue->WorkGroupPartitioning(armcl_kernel,
                    work_dim,
                    global_work_offset,
                    global_work_size,
                    local_work_size,
                    num_events_in_wait_list,
                    event_wait_list,
                    event);
        } else if (!strncmp("OnTheFly", option, strlen(option))) {
            printf("this is OnTheFly option\n");

            return armcl_command_queue->OnTheFly(armcl_kernel,
                    work_dim,
                    global_work_offset,
                    global_work_size,
                    local_work_size,
                    num_events_in_wait_list,
                    event_wait_list,
                    event);
        } else if (!strncmp("Manual", option, strlen(option))) {
            printf("this is Manual partitioning option\n");

            // compute array_cpu and array_gpu once
            return armcl_command_queue->Manual_Static(armcl_kernel,
                    work_dim,
                    global_work_offset,
                    global_work_size,
                    local_work_size,
                    num_events_in_wait_list,
                    event_wait_list,
                    event);
        } else {
            printf("not supported option\n");
        }
    }

    std::cout << "Normal execution" << std::endl;
    return clEnqueueNDRangeKernel_p(command_queue,
            kernel,
            work_dim,
            global_work_offset,
            global_work_size,
            local_work_size,
            num_events_in_wait_list,
            event_wait_list,
            event);
}

