//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) Computer Systems and Platforms Laboratory
// http://csap.snu.ac.kr
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLCommandQueue.h"
#include <string.h>
#include <algorithm>
#include <time.h>

int ArmCLCommandQueue::CPUThreading(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "CPUThreading" << std::endl;

    const char* kernel_name = armcl_kernel->kernel_name();
    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();

    struct timespec start, end;
    long long elapsed_time_us;

    int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

    int (*mbap_cpu_thread_kernel)(int id, void* argp);
    mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

    int mbap_cpu_thread_kernel_num_args =
        armcl_kernel->mbap_cpu_thread_kernel_num_args();
    unsigned long* mbap_cpu_thread_kernel_params =
        armcl_kernel->mbap_cpu_thread_kernel_params();

    // register reordered map for CPU threads
    for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
        mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
            + mbap_cpu_thread_kernel_num_args - 5] =
            (unsigned long)in_order_map_;
    }

    clock_gettime(CLOCK_REALTIME, &start);

    std::vector<std::thread*> threads;

    size_t alloc_wgs = total_num_wgs / num_cpu_cores;
    size_t alloc_index = 0;

    /* equal distribution */
    for (int i = 0; i < num_cpu_cores; i++) {
        if (i == num_cpu_cores-1) {
            // register work item reordered map and
            // work group partitioning information
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 4] =
                (unsigned long)alloc_index;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 3] =
                (unsigned long)local_work_size[0];
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 2] =
                (unsigned long)total_num_wgs - alloc_index;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 1] =
                (unsigned long)(&has_finished[i]);

            std::thread* thread_ = new std::thread(
                    mbap_cpu_thread_kernel,
                    i,
                    (void*)NULL);
            threads.push_back(thread_);
        } else {
            // register work item reordered map and
            // work group partitioning information
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 4] =
                (unsigned long)alloc_index;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 3] =
                (unsigned long)local_work_size[0];
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 2] =
                (unsigned long)alloc_wgs;
            mbap_cpu_thread_kernel_params[
                i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 1] =
                (unsigned long)(&has_finished[i]);

            std::thread* thread_ = new std::thread(
                    mbap_cpu_thread_kernel,
                    i,
                    (void*)NULL);
            threads.push_back(thread_);
        }

        alloc_index += alloc_wgs;
    }

    for (int i = 0; i < num_cpu_cores; i++) {
        threads[i]->join();
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;

    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}

