//------------------------------------------------------------------------------
// ARM OpenCL Runtime
//
// Copyright (c) Computer Systems and Platforms Laboratory
// http://csap.snu.ac.kr
// All right reserved.
//
// Author:
// Younghyun Cho <younghyun@csap.snu.ac.kr>
//

#include "ArmCLCommandQueue.h"
#include <string.h>
#include <algorithm>
#include <time.h>

int ArmCLCommandQueue::WorkGroupPartitioning(ArmCLKernel* armcl_kernel,
        cl_uint work_dim,
        const size_t* global_work_offset,
        const size_t* global_work_size,
        const size_t* local_work_size,
        cl_uint num_events_in_wait_list,
        const cl_event* event_wait_list,
        cl_event* event)
{
    std::cout << "Work group partitioning (Manual)" << std::endl;
    const char* kernel_name = armcl_kernel->kernel_name();

    size_t total_work_size = global_work_size[0];
    size_t total_num_wgs = total_work_size / local_work_size[0];
    size_t wg_size = local_work_size[0];
    //std::cout << "total_work_size: " << total_work_size << std::endl;
    //std::cout << "total number of work groups: " << total_num_wgs << std::endl;

    int num_gpu_cores = armcl_kernel->num_proper_gpu_cores();
    int num_cpu_cores = armcl_kernel->num_proper_cpu_cores();
    //std::cout << "num_gpu_cores: " << num_gpu_cores << std::endl;
    //std::cout << "num_cpu_cores: " << num_cpu_cores << std::endl;

    struct timespec start, end;
    long long elapsed_time_us;

    cl_int err;

    int gpu_ratio = atoi(std::getenv("GPU_RATIO"));
    int cpu_ratio = atoi(std::getenv("CPU_RATIO"));
    //std::cout << "gpu_ratio: " << gpu_ratio << std::endl;
    //std::cout << "cpu_ratio: " << cpu_ratio << std::endl;

    size_t gpu_wgs = (size_t)((double)total_num_wgs * ((double)gpu_ratio / 100.0));
    size_t cpu_wgs = total_num_wgs - gpu_wgs;
    //std::cout << "number of work groups for GPU: " << gpu_wgs << std::endl;
    //std::cout << "number of work groups for CPU: " << cpu_wgs << std::endl;

    size_t gpu_partitioned_global_work_size[1];
    gpu_partitioned_global_work_size[0] = gpu_wgs * local_work_size[0];
    size_t cpu_partitioned_global_work_size[1];
    cpu_partitioned_global_work_size[0] = cpu_wgs * local_work_size[0];

    cl_mem m_in_order_map = clCreateBuffer_p_(context_,
            CL_MEM_READ_ONLY|CL_MEM_USE_HOST_PTR,
            total_work_size * sizeof(int),
            in_order_map_,
            &err);

    clock_gettime(CLOCK_REALTIME, &start);
    // execution
    if (gpu_wgs) {
        cl_kernel mbap_kernel = armcl_kernel->mbap_kernel();
        cl_int mbap_kernel_num_args = armcl_kernel->mbap_kernel_num_args();

        err = clSetKernelArg_p_(mbap_kernel, mbap_kernel_num_args-1, sizeof(cl_mem), (void*)&m_in_order_map);
        if (err != CL_SUCCESS) {
            std::cout << "ArmCL: clSetKernelArg for mbap_kernel" << std::endl;
        }

        err = clEnqueueNDRangeKernel_p_(command_queue_,
                mbap_kernel,
                work_dim,
                global_work_offset,
                gpu_partitioned_global_work_size,
                local_work_size,
                num_events_in_wait_list,
                event_wait_list,
                event);
    }

    if (cpu_wgs) {
        int* has_finished = (int*)calloc(num_cpu_cores, sizeof(int));

        int (*mbap_cpu_thread_kernel)(int id, void* argp);
        mbap_cpu_thread_kernel = armcl_kernel->mbap_cpu_thread_kernel();

        int mbap_cpu_thread_kernel_num_args =
            armcl_kernel->mbap_cpu_thread_kernel_num_args();
        unsigned long* mbap_cpu_thread_kernel_params =
            armcl_kernel->mbap_cpu_thread_kernel_params();

        // register reordered map for CPU threads
        for (int i = 0; i < armcl_context_->num_cpu_compute_units(); i++) {
            mbap_cpu_thread_kernel_params[i*mbap_cpu_thread_kernel_num_args
                + mbap_cpu_thread_kernel_num_args - 5] =
                (unsigned long)in_order_map_;
        }

        std::vector<std::thread*> threads;

        size_t alloc_wgs = cpu_wgs / num_cpu_cores;
        size_t alloc_index = gpu_wgs;

        for (int i = 0; i < num_cpu_cores; i++) {
            std::cout << "alloc: " << i << " alloc_wg: " << alloc_index
                << " alloc_num_wgs: " << alloc_wgs << std::endl;

            if (i == num_cpu_cores-1) {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)in_order_map_;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)total_num_wgs - alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            } else {
                // register work item reordered map and
                // work group partitioning information
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 5] =
                    (unsigned long)in_order_map_;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 4] =
                    (unsigned long)alloc_index;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 3] =
                    (unsigned long)local_work_size[0];
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 2] =
                    (unsigned long)alloc_wgs;
                mbap_cpu_thread_kernel_params[
                    i*mbap_cpu_thread_kernel_num_args
                    + mbap_cpu_thread_kernel_num_args - 1] =
                    (unsigned long)(&has_finished[i]);

                std::thread* thread_ = new std::thread(
                        mbap_cpu_thread_kernel,
                        i,
                        (void*)NULL);
                threads.push_back(thread_);
            }

            alloc_index += alloc_wgs;
        }

        for (int i = 0; i < num_cpu_cores; i++) {
            threads[i]->join();
        }
    }

    err = clFinish(command_queue_);
    if (err != CL_SUCCESS) {
        std::cout << "ArmCL MBAP: "
            << "clFinish error"
            << std::endl;
    }

    clock_gettime(CLOCK_REALTIME, &end);
    elapsed_time_us =
        ((end.tv_sec - start.tv_sec)*1000000000 + (end.tv_nsec - start.tv_nsec))/1000;
    std::cout << kernel_name << " kernel elapsed time: " << elapsed_time_us << std::endl;

    return CL_SUCCESS;
}


